"""Optical lever sensing matrix
"""
import numpy as np
import scipy

import kontrol


class OpticalLeverSensingMatrix(kontrol.SensingMatrix):  # Too long?
    r"""Optical lever sensing matrix base class.

    This is a sensing matrix that maps tilt-sensing QPD and
    length-sensing QPD (placed behind a lens) readouts to
    the suspended optics' longitudinal, pitch, and yaw displacements.

    Parameters
    ----------
    r_h: float
        Lever arm from the optics to the tilt-sensing QPD plane on the
        horizontal plane (amplifying yaw).
    r_v: float
        Lever arm from the optics to the tilt-sensing QPD plane on the
        vertical plane (amplifying pitch).
    alpha_h: float
        Angle of incidence on the horizontal plane.
    alpha_v: float
        Angle of incidence on the vertical plane.
    r_lens_h: float, optional
        Lever arm from optics to the lens on the horizontal plane.
        Defaults None. Specify if it exists.
    r_lens_v: float, optional
        Lever arm from optics to the lens on the vertical plane.
        Defaults None. Specify if it exists.
    d_h: float, optional
        Horizontal distance from the lens to the length-sensing QPD.
        Defaults None.
    d_v: float, optional
        Vertical distance from the lens to the length-sensing QPD.
        Defaults None.
    delta_x: float, optional
        Horizontal miscentering of the beam spot at the optics plane.
    delta_y: float, optional
        Vertical miscentering of the beam spot at the optics plane.
    phi_tilt: float, optional
        Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
        Defaults None.
    phi_len: float, optional
        Angle from the length-sensing QPD frame to the yaw-pitch frame.
        Defaults None.
    f_h : float, optional,
        Focal length of the lens projected to the horizontal plane.
        Defaults ``np.inf`` (no lens).
    f_v : float, optional,
        Focal length of the lens projected to the vertical plane.
        Defaults ``np.inf`` (no lens).
    format: str, optional
        Format of the sensing matrix.
        Choose from
            "OPLEV2EUL": Default sensing matrix from KAGRA MEDM screen
                with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                and output (longitudinal, pitch and yaw).
            "xy": Matrix as shown in [1]_.
    coupling_matrix: array, optional
        The coupling matrix.
        Default None.

    Notes
    -----
    We're using equation (29) from [1]_.

    .. math::
        \begin{pmatrix}
            x_L\\
            \theta_P\\
            \theta_Y
        \end{pmatrix}
        =
        \mathbf{C}_\mathrm{miscenter}
        \mathbf{C}_\mathrm{align}
        \mathbf{C}_\mathrm{rotation}
        \begin{pmatrix}
            x_\mathrm{tilt}\\
            y_\mathrm{tilt}\\
            x_\mathrm{len}\\
            y_\mathrm{len}
        \end{pmatrix},

    where :math:`x_L` is the longitudinal displacement of the optics,
    :math:`\theta_P` is the pitch angular displacement of the optics,
    :math:`\theta_Y` is the yaw angular displacement of the optics,
    :math:`x_\mathrm{tilt}` is the horizontal displacement of the beam spot
    at the tilt-sensing QPD plane,
    :math:`y_\mathrm{tilt}` is the vertical displacement of the beam spot
    at the tilt-sensing QPD plane,
    :math:`x_\mathrm{len}` is the horizontal displacement of the beam spot
    at the length-sensing QPD plane,
    :math:`y_\mathrm{len}` is the vertical displacement of the beam spot
    at the length-sensing QPD plane,

    .. math::
        \mathbf{C}_\mathrm{rotation}
        =
        \begin{bmatrix}
            \cos\phi_\mathrm{tilt} & \sin\phi_\mathrm{tilt} & 0 & 0\\
            -\sin\phi_\mathrm{tilt} & \cos\phi_\mathrm{tilt} & 0 & 0\\
            0 & 0 & \cos\phi_\mathrm{len} & \sin\phi_\mathrm{len}\\
            0 & 0 & -\sin\phi_\mathrm{len} & \cos\phi_\mathrm{len}
        \end{bmatrix},

    .. math::
        \mathbf{C}_\mathrm{align}
        =
        \begin{bmatrix}
            2\sin\alpha_h & 0 & 2r_h\\
            2\sin\alpha_v & 2r_v & 0\\
            2\sin\alpha_h\left(1-\frac{d_h}{f_h}\right) & 0 &
            2\left[\left(1-\frac{d_h}{f_h}\right)r_{\mathrm{lens},h}
            + d_h\right]\\
            2\sin\alpha_v\left(1-\frac{d_v}{f_v}\right) &
            2\left[\left(1-\frac{d_v}{f_v}\right)r_{\mathrm{lens},v}
            + d_v\right]
            & 0\\
        \end{bmatrix}^{+},

    .. math::
        \mathbf{C}_\mathrm{miscenter}
        =
        \begin{bmatrix}
            1 & \delta_y & \delta_x\\
            0 & 1 & 0\\
            0 & 0 & 1
        \end{bmatrix}^{-1},

    :math:`\phi_\mathrm{tilt}` is the angle between the tilt-sensing QPD
    and the yaw-pitch frame, :math:`\phi_\mathrm{len}` is the angle between
    the length-sensing QPD and the yaw-pitch frame,
    :math:`r_h` is the lever arm on the horzontal plane, :math:`r_v` is the
    lever arm on the vertical plane,
    :math:`\alpha_h` is the angle of incidence on the horizontal plane,
    :math:`\alpha_v` is the angle of incidence on the vertical plane,
    :math:`r_{\mathrm{lens}, h}` is the lever arm between the optics
    and the lens on the horizontal plane,
    :math:`r_{\mathrm{lens}, v}` is the lever arm between the optics and the
    lens on the vertical plane,
    :math:`d_h` is the distance between the lens and the length-sensing QPD
    on the horizontal plane,
    :math:`d_v` is the distance between lens and the length-sensing QPD
    on the vertical plane,
    and :math:`f` is the focal length of the convex lens.

    References
    ----------
    .. [1]
        Tsang Terrence Tak Lun,
        Sensing Matrices for Optical Levers of the KAGRA Main Optics,
        https://github.com/terrencetec/kagra-optical-lever.
    """
    def __new__(cls, r_h, r_v, alpha_h, alpha_v,
                r_lens_h=0, r_lens_v=0, d_h=0, d_v=0,
                delta_x=0, delta_y=0,
                phi_tilt=0, phi_len=0, f_h=np.inf, f_v=np.inf,
                format="OPLEV2EUL",
                coupling_matrix=None, *args, **kwargs):
        r"""Constructor

        Parameters
        ----------
        r_h: float
            Lever arm from the optics to the tilt-sensing QPD plane on the
            horizontal plane (amplifying yaw).
        r_v: float
            Lever arm from the optics to the tilt-sensing QPD plane on the
            vertical plane (amplifying pitch).
        alpha_h: float
            Angle of incidence on the horizontal plane.
        alpha_v: float
            Angle of incidence on the vertical plane.
        r_lens_h: float, optional
            Lever arm from optics to the lens on the horizontal plane.
            Defaults 0.
        r_lens_v: float, optional
            Lever arm from optics to the lens on the vertical plane.
            Defaults 0.
        d_h: float, optional
            Horizontal distance from the lens to the length-sensing QPD.
            Defaults 0.
        d_v: float, optional
            Vertical distance from the lens to the length-sensing QPD.
            Defaults 0.
        delta_x: float, optional
            Horizontal miscentering of the beam spot at the optics plane.
            Defaults 0.
        delta_y: float, optional
            Vertical miscentering of the beam spot at the optics plane.
            Defaults 0.
        phi_tilt: float, optional
            Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
            Defaults 0.
        phi_len: float, optional
            Angle from the length-sensing QPD frame to the yaw-pitch frame.
            Defaults 0.
        f_h : float, optional,
            Focal length of the lens projected to the horizontal plane.
            Defaults ``np.inf`` (no lens).
        f_v : float, optional,
            Focal length of the lens projected to the vertical plane.
            Defaults ``np.inf`` (no lens).
        format: str, optional
            Format of the sensing matrix.
            Choose from
                "OPLEV2EUL": Default sensing matrix from KAGRA MEDM screen
                    with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                    and output (longitudinal, pitch and yaw).
                "xy": Matrix as shown in [1]_.
        coupling_matrix: array, optional
            The coupling matrix.
            Default None.

        Notes
        -----
        The coupling matrix has (i, j) elements as coupling ratios x_i/x_j.
        For example, consider the 2-sensor configuration:
        I have a coupled sensing readout :math:`x_{1,\mathrm{coupled}}`
        that reads :math:`x_{1,\mathrm{coupled}}=x_1 + 0.1x_2`.
        and, I have another coupled sensing readout
        :math:`x_{2,\mathrm{coupled}}` that reads
        :math:`x_{2,\mathrm{coupled}}=-0.2x_1 + x_2`.
        Then, the coupling matrix is

        .. math::
            \begin{bmatrix}
            1 & 0.1\\
            -0.2 & 1
            \end{bmatrix}.

        References
        ----------
        .. [1]
            Tsang Terrence Tak Lun,
            Sensing Matrices for Optical Levers of the KAGRA Main Optics,
            https://github.com/terrencetec/kagra-optical-lever.
        """
        # print("OpticalLeverSensingMatrix __new__")
        _c_align = c_align(
            r_h=r_h, r_v=r_v, alpha_h=alpha_h, alpha_v=alpha_v,
            r_lens_h=r_lens_h, r_lens_v=r_lens_v, d_h=d_h, d_v=d_v,
            f_h=f_h, f_v=f_v,
            **kwargs)
        _c_rotation = c_rotation(phi_tilt, phi_len)
        _c_miscenter = c_miscenter(delta_x, delta_y)

        c_sensing = _c_miscenter @ _c_align @ _c_rotation

        self = super(OpticalLeverSensingMatrix, cls).__new__(
            cls, matrix=c_sensing)  # __new__ in kontrol.sensact.matrix.Matrix
        super().__init__(self, matrix=None, coupling_matrix=coupling_matrix)
        self.c_align = _c_align
        self.c_rotation = _c_rotation
        self.c_miscenter = _c_miscenter
        self.c_sensing = c_sensing
        self.r_h = r_h
        self.r_v = r_v
        self.alpha_h = alpha_h
        self.alpha_v = alpha_v
        self.r_lens_h = r_lens_h
        self.r_lens_v = r_lens_v
        self.d_h = d_h
        self.d_v = d_v
        self.delta_x = delta_x
        self.delta_y = delta_y
        self.phi_tilt = phi_tilt
        self.phi_len = phi_len
        self.f_h = f_h
        self.f_v = f_v
        self.format = format
        return self

    def __init__(self, *args, **kwargs):
        """Constructor
        """
        # print("OpticalLeverSensingMatrix __init__")
        # __init__ in kontrol.sensact.matrix.SensingMatrix
        # super().__init__(matrix=None, coupling_matrix=coupling_matrix)
        pass

    def update_matrices_decorator(func):
        """Update matrices and self upon setting new parameters.
        """
        def update_matrices(self, *args, **kwargs):
            func(self, *args, **kwargs)
            try:
                # print("In decorator")
                _c_align = c_align(
                    r_h=self.r_h, r_v=self.r_v,
                    alpha_h=self.alpha_h, alpha_v=self.alpha_v,
                    r_lens_h=self.r_lens_h, r_lens_v=self.r_lens_v,
                    d_h=self.d_h, d_v=self.d_v, f_h=self.f_h, f_v=self.f_v,
                    **kwargs)
                _c_rotation = c_rotation(self.phi_tilt, self.phi_len)
                _c_miscenter = c_miscenter(self.delta_x, self.delta_y)
                c_sensing = _c_miscenter @ _c_align @ _c_rotation
                if self.format in ["OPLEV2EUL", "OL2EUL"]:
                    c_sensing[:, [0, 1, 2, 3]] = c_sensing[:, [1, 0, 3, 2]]
                np.copyto(self, c_sensing)
            except AttributeError:
                pass
        return update_matrices

    @property
    def r_h(self):
        """Lever arm from optics to tilt-sensing QPD on the horizontal plane.
        """
        return self._r_h

    @r_h.setter
    @update_matrices_decorator
    def r_h(self, _r_h):
        """r_h setter.
        """
        self._r_h = _r_h

    @property
    def r_v(self):
        """Lever arm from optics to tilt-sensing QPD on the vertical plane.
        """
        return self._r_v

    @r_v.setter
    @update_matrices_decorator
    def r_v(self, _r_v):
        """r_v setter
        """
        self._r_v = _r_v

    @property
    def alpha_h(self):
        """Angle of incidence on the horizontal plane.
        """
        return self._alpha_h

    @alpha_h.setter
    @update_matrices_decorator
    def alpha_h(self, _alpha_h):
        """alpha_h setter
        """
        self._alpha_h = _alpha_h

    @property
    def alpha_v(self):
        """Angle of incidence on the vertical plane.
        """
        return self._alpha_v

    @alpha_v.setter
    @update_matrices_decorator
    def alpha_v(self, _alpha_v):
        """alpha_v setter
        """
        self._alpha_v = _alpha_v

    @property
    def r_lens_h(self):
        """Lever arm from optics to the lens on the horizontal plane.
        """
        return self._r_lens_h

    @r_lens_h.setter
    @update_matrices_decorator
    def r_lens_h(self, _r_lens_h):
        """r_lens_h setter
        """
        self._r_lens_h = _r_lens_h

    @property
    def r_lens_v(self):
        """Lever arm from optics to the lens on the vertical plane.
        """
        return self._r_lens_v

    @r_lens_v.setter
    @update_matrices_decorator
    def r_lens_v(self, _r_lens_v):
        """r_lens_v setter
        """
        self._r_lens_v = _r_lens_v

    @property
    def d_h(self):
        """Horizontal distance from the lens to the length-sensing QPD.
        """
        return self._d_h

    @d_h.setter
    @update_matrices_decorator
    def d_h(self, _d_h):
        """d_h setter
        """
        self._d_h = _d_h

    @property
    def d_v(self):
        """Vertical distance from the lens to the length-sensing QPD.
        """
        return self._d_v

    @d_v.setter
    @update_matrices_decorator
    def d_v(self, _d_v):
        """d_v setter
        """
        self._d_v = _d_v

    @property
    def delta_x(self):
        """Horizontal miscentering of the beam spot at the optics plane.
        """
        return self._delta_x

    @delta_x.setter
    @update_matrices_decorator
    def delta_x(self, _delta_x):
        """delta_x setter
        """
        self._delta_x = _delta_x

    @property
    def delta_y(self):
        """Horizontal miscentering of the beam spot at the optics plane.
        """
        return self._delta_y

    @delta_y.setter
    @update_matrices_decorator
    def delta_y(self, _delta_y):
        """delta_y setter
        """
        self._delta_y = _delta_y

    @property
    def phi_tilt(self):
        """Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
        """
        return self._phi_tilt

    @phi_tilt.setter
    @update_matrices_decorator
    def phi_tilt(self, _phi_tilt):
        """phi_tilt setter
        """
        self._phi_tilt = _phi_tilt

    @property
    def phi_len(self):
        """Angle from the length-sensing QPD frame to the yaw-pitch frame.
        """
        return self._phi_len

    @phi_len.setter
    @update_matrices_decorator
    def phi_len(self, _phi_len):
        """phi_len setter
        """
        self._phi_len = _phi_len

    @property
    def f_h(self):
        """Focal length of the convex lens projected on the horizontal plane.
        """
        return self._f_h

    @f_h.setter
    @update_matrices_decorator
    def f_h(self, _f_h):
        """f_h.setter
        """
        self._f_h = _f_h

    @property
    def f_v(self):
        """Focal length of the convex lens projected on the vertical plane.
        """
        return self._f_v

    @f_v.setter
    @update_matrices_decorator
    def f_v(self, _f_v):
        """f_v.setter
        """
        self._f_v = _f_v

    @property
    def format(self):
        """Format of the sensing matrix.

        Choose from
            "OPLEV2EUL": Default sensing matrix from KAGRA MEDM screen
                with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                and output (longitudinal, pitch and yaw).
            "xy": Matrix as shown in [1]_.
        References
        ----------
        .. [1]
            Tsang Terrence Tak Lun,
            Sensing Matrices for Optical Levers of the KAGRA Main Optics,
            https://github.com/terrencetec/kagra-optical-lever.
        """
        return self._format

    @format.setter
    @update_matrices_decorator
    def format(self, _format):
        """format setter. Format of the sensing matrix.

        Parameters
        ----------
        _format: str
            Choose from
                "OPLEV2EUL"/"OP2EUL": Default sensing matrix from KAGRA MEDM
                    screen with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                    and output (longitudinal, pitch and yaw).
                "xy": Matrix as shown in [1]_.

        References
        ----------
        .. [1]
            Tsang Terrence Tak Lun,
            Sensing Matrices for Optical Levers of the KAGRA Main Optics,
            https://github.com/terrencetec/kagra-optical-lever.
        """
        if _format not in ["OPLEV2EUL", "OL2EUL", "xy"]:
            raise ValueError("Choose from [\"OPLEV2EUL\", \"OL2EUL\", \"xy\"]")
        self._format = _format


class HorizontalOpticalLeverSensingMatrix(OpticalLeverSensingMatrix):
    """Horizontal optical lever sensing matrix.

    Parameters
    ----------
    r: float
        Lever arm.
    alpha_h: float
        Angle of incidence on the horizontal plane.
    r_lens: float, optional
        Lever arm from the optics to the convex lens.
        Default 0.
    f: float, optional
        Focal length of the convex lens.
        Default np.inf.
    alpha_v: float, optional
        Angle of incidence on the vertical plane.
        Default 0.
    phi_tilt: float, optional
        Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
        Default 0.
    phi_len: float, optional
        Angle from the length-sensing QPD frame to the yaw-pitch frame.
        Default 0.
    delta_x: float, optional
        Horizontal miscentering of the beam spot at the optics plane.
        Default 0.
    delta_y: float, optional
        Vertical miscentering of the beam spot at the optics plane.
        Default 0.
    delta_d: float, optional
        Misplacement of the length-sensing QPD.
        Default 0.
    format: str, optional
        Format of the sensing matrix.
        Choose from
            "OPLEV2EUL": Default sensing matrix from KAGRA MEDM screen
                with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                and output (longitudinal, pitch and yaw).
            "xy": Matrix as shown in [1]_.
    coupling_matrix: array, optional
        The coupling matrix.
        Default None.
    *args:
        Variable length arguments passed to OpticalLeverSensingMatrix.
    **kwargs:
        Keyword arguments passed to OpticalLeverSensingMatrix.
    """
    def __new__(cls, r, alpha_h, r_lens=0, f=np.inf,
                alpha_v=0, phi_tilt=0, phi_len=0,
                delta_x=0, delta_y=0, delta_d=0,
                format="OPLEV2EUL", coupling_matrix=None, *args, **kwargs):
        """Constructor

        Parameters
        ----------
        r: float
            Lever arm.
        alpha_h: float
            Angle of incidence on the horizontal plane.
        r_lens: float, optional
            Lever arm from the optics to the convex lens.
            Default 0.
        f: float, optional
            Focal length of the convex lens.
            Default np.inf.
        alpha_v: float, optional
            Angle of incidence on the vertical plane.
            Default 0.
        phi_tilt: float, optional
            Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
            Default 0.
        phi_len: float, optional
            Angle from the length-sensing QPD frame to the yaw-pitch frame.
            Default 0.
        delta_x: float, optional
            Horizontal miscentering of the beam spot at the optics plane.
            Default 0.
        delta_y: float, optional
            Vertical miscentering of the beam spot at the optics plane.
            Default 0.
        delta_d: float, optional
            Misplacement of the length-sensing QPD.
            Default 0.
        format: str, optional
            Format of the sensing matrix.
            Choose from
                "OPLEV2EUL": Default sensing matrix from KAGRA MEDM screen
                    with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                    and output (longitudinal, pitch and yaw).
                "xy": Matrix as shown in [1]_.
        coupling_matrix: array, optional
            The coupling matrix.
            Default None.
        *args:
            Variable length arguments passed to OpticalLeverSensingMatrix.
        **kwargs:
            Keyword arguments passed to OpticalLeverSensingMatrix.
        """
        r_h = r
        r_v = r*np.cos(alpha_h)
        r_lens_h = r_lens
        if f is not np.inf:
            d_h = r_lens*f / (r_lens-f) + delta_d
            f_h = f
            f_v = r*np.cos(alpha_h)
        else:
            d_h = 0
            f_h = np.inf
            f_v = np.inf
        self = super(HorizontalOpticalLeverSensingMatrix, cls).__new__(
            cls,
            r_h=r_h, r_v=r_v, alpha_h=alpha_h, alpha_v=alpha_v,
            r_lens_h=r_lens_h, f_h=f_h, f_v=f_v, d_h=d_h,
            phi_tilt=phi_tilt, phi_len=phi_len,
            delta_x=delta_x, delta_y=delta_y,
            format=format, coupling_matrix=coupling_matrix, *args, **kwargs
        )
        self.r = r
        self.r_lens = r_lens
        self.delta_d = delta_d
        return self

    def __init__(self, *args, **kwargs):
        """Constructor
        """
        pass

    @property
    def r(self):
        """Lever arm.
        """
        return self._r

    @r.setter
    # @OpticalLeverSensingMatrix.update_matrices_decorator
    def r(self, _r):
        """r setter
        """
        self._r = _r
        self.r_h = self.r
        self.r_v = self.r * np.cos(self.alpha_h)

    @property
    def r_lens(self):
        """Lever arm from the optics to the convex lens.
        """
        return self._r_lens

    @r_lens.setter
    def r_lens(self, _r_lens):
        """r_lens setter
        """
        self._r_lens = _r_lens
        self.r_lens_h = self.r_lens

    @property
    def delta_d(self):
        """Misplacement of the length-sensing QPD.
        """
        return self._delta_d

    @delta_d.setter
    def delta_d(self, _delta_d):
        """delta_d setter
        """
        self._delta_d = _delta_d
        self.d_h = self.r_lens*self.f_h/(self.r_lens-self.f_h) + self._delta_d

    @property
    def f(self):
        """Focal length of the convex lens.
        """
        return self._f

    @f.setter
    def f(self, _f):
        """f setter
        """
        self._f = _f
        self.f_h = _f
        self.f_v = _f * np.cos(self.alpha_h)


class VerticalOpticalLeverSensingMatrix(OpticalLeverSensingMatrix):
    """Vertical optical lever sensing matrix.

    Parameters
    ----------
    r: float
        Lever arm.
    alpha_v: float
        Angle of incidence on the vertical plane.
    r_lens: float, optional
        Lever arm from the optics to the convex lens.
        Default 0.
    f: float, optional
        Focal length of the convex lens.
        Default np.inf.
    alpha_h: float, optional
        Angle of incidence on the horizontal plane.
        Default 0.
    phi_tilt: float, optional
        Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
        Default 0.
    phi_len: float, optional
        Angle from the length-sensing QPD frame to the yaw-pitch frame.
        Default 0.
    delta_x: float, optional
        Horizontal miscentering of the beam spot at the optics plane.
        Default 0.
    delta_y: float, optional
        Vertical miscentering of the beam spot at the optics plane.
        Default 0.
    delta_d: float, optional
        Misplacement of the length-sensing QPD.
        Default 0.
    format: str, optional
        Format of the sensing matrix.
        Choose from
            "OPLEV2EUL": Default sensing matrix from KAGRA MEDM screen
                with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                and output (longitudinal, pitch and yaw).
            "xy": Matrix as shown in [1]_.
    coupling_matrix: array, optional
        The coupling matrix.
        Default None.
    *args:
        Variable length arguments passed to OpticalLeverSensingMatrix.
    **kwargs:
        Keyword arguments passed to OpticalLeverSensingMatrix.
    """
    def __new__(cls, r, alpha_v, r_lens=0, f=np.inf,
                alpha_h=0, phi_tilt=0, phi_len=0,
                delta_x=0, delta_y=0, delta_d=0,
                format="OPLEV2EUL", coupling_matrix=None, *args, **kwargs):
        """Constructor

        Parameters
        ----------
        r: float
            Lever arm.
        alpha_v: float
            Angle of incidence on the vertical plane.
        r_lens: float, optional
            Lever arm from the optics to the convex lens.
            Default 0.
        f: float, optional
            Focal length of the convex lens.
            Default np.inf.
        alpha_h: float, optional
            Angle of incidence on the horizontal plane.
            Default 0.
        phi_tilt: float, optional
            Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
            Default 0.
        phi_len: float, optional
            Angle from the length-sensing QPD frame to the yaw-pitch frame.
            Default 0.
        delta_x: float, optional
            Horizontal miscentering of the beam spot at the optics plane.
            Default 0.
        delta_y: float, optional
            Vertical miscentering of the beam spot at the optics plane.
            Default 0.
        delta_d: float, optional
            Misplacement of the length-sensing QPD.
            Default 0.
        format: str, optional
            Format of the sensing matrix.
            Choose from
                "OPLEV2EUL": Default sensing matrix from KAGRA MEDM screen
                    with input (TILT_PIT, TILT_YAW, LEN_PIT, LEN_YAW),
                    and output (longitudinal, pitch and yaw).
                "xy": Matrix as shown in [1]_.
        coupling_matrix: array, optional
            The coupling matrix.
            Default None.
        *args:
            Variable length arguments passed to OpticalLeverSensingMatrix.
        **kwargs:
            Keyword arguments passed to OpticalLeverSensingMatrix.
        """
        r_h = r*np.cos(alpha_v)
        r_v = r
        r_lens_v = r_lens
        if f is not np.inf:
            d_v = r_lens*f / (r_lens-f) + delta_d
            f_h = f*np.cos(alpha_v)
            f_v = f
        else:
            d_v = 0
            f_h = np.inf
            f_v = np.inf
        self = super(VerticalOpticalLeverSensingMatrix, cls).__new__(
            cls,
            r_h=r_h, r_v=r_v, alpha_h=alpha_h, alpha_v=alpha_v,
            r_lens_v=r_lens_v, f_h=f_h, f_v=f_v, d_v=d_v,
            phi_tilt=phi_tilt, phi_len=phi_len,
            delta_x=delta_x, delta_y=delta_y,
            format=format, coupling_matrix=coupling_matrix, *args, **kwargs
        )
        self.r = r
        self.r_lens = r_lens
        self.delta_d = delta_d
        self.f = f
        return self

    def __init__(self, *args, **kwargs):
        """Constructor
        """
        pass

    @property
    def r(self):
        """Lever arm.
        """
        return self._r

    @r.setter
    # @OpticalLeverSensingMatrix.update_matrices_decorator
    def r(self, _r):
        """r setter
        """
        self._r = _r
        self.r_h = self.r * np.cos(self.alpha_v)
        self.r_v = self.r

    @property
    def r_lens(self):
        """Lever arm from the optics to the convex lens.
        """
        return self._r_lens

    @r_lens.setter
    def r_lens(self, _r_lens):
        """r_lens setter
        """
        self._r_lens = _r_lens
        self.r_lens_v = self.r_lens

    @property
    def delta_d(self):
        """Misplacement of the length-sensing QPD.
        """
        return self._delta_d

    @delta_d.setter
    def delta_d(self, _delta_d):
        """delta_d setter
        """
        self._delta_d = _delta_d
        self.d_v = self.r_lens*self.f_v/(self.r_lens-self.f_v) + self._delta_d

    @property
    def f(self):
        """Focal length of the convex lens.
        """
        return self._f

    @f.setter
    def f(self, _f):
        """f setter
        """
        self._f = _f
        self.f_h = _f * np.cos(self.alpha_v)
        self.f_v = _f


def c_align(r_h, r_v, alpha_h, alpha_v,
            r_lens_h=0, r_lens_v=0, d_h=0, d_v=0,
            f_h=np.inf, f_v=np.inf,
            roundoff=6):
    r"""Return optical lever sensing matrix for a perfectly aligned case.

    Parameters
    ----------
    r_h: float
        Lever arm from the optics to the tilt-sensing QPD plane on the
        horizontal plane (amplifying yaw).
    r_v: float
        Lever arm from the optics to the tilt-sensing QPD plane on the
        vertical plane (amplifying pitch).
    alpha_h: float
        Angle of incidence on the horizontal plane.
    alpha_v: float
        Angle of incidence on the vertical plane.
    r_lens_h: float, optional
        Lever arm from optics to the lens on the horizontal plane.
        Defaults None. Specify if it exists.
    r_lens_v: float, optional
        Lever arm from optics to the lens on the vertical plane.
        Defaults None. Specify if it exists.
    d_h: float, optional
        Horizontal distance from the lens to the length-sensing QPD.
        Defaults None.
    d_v: float, optional
        Vertical distance from the lens to the length-sensing QPD.
        Defaults None.
    f_h : float, optional,
        Focal length of the lens projected to the horizontal plane.
        Defaults ``np.inf`` (no lens).
    f_v : float, optional,
        Focal length of the lens projected to the vertical plane.
        Defaults ``np.inf`` (no lens).
    roundoff: int, optional
        How many decimal places to keep.

    Returns
    -------
    array
        The aligned optical lever sensing matrix

    Notes
    -----
    See :math:`\mathbf{C}_\mathrm{align}` from [1]_.

    If `f_h` or `f_v` or (`d_h` and `d_v`) or (`r_lens_h` and `r_lens_v`)
    are not  specified, then we assume no length-sensing QPD so first column,
    third row, and forth row will be 0.
    elif `f_h` and `f_v` is specified, (`d_h` or `r_lens_h`) are not specified,
    then third row will be 0.
    elif `f_h` and `f_v` is specified, (`d_v` or `r_lens_v`) are not specified,
    then forth row will be 0.

    References
    ----------
    .. [1]
        Tsang Terrence Tak Lun,
        Sensing Matrices for Optical Levers of the KAGRA Main Optics,
        https://github.com/terrencetec/kagra-optical-lever.
    """
    # print(r_h, r_v, alpha_h, alpha_v, r_lens_h, r_lens_v, d_h, d_v, f)
    c_align_inv = np.array([
            [2*np.sin(alpha_h), 0, 2*r_h],
            [2*np.sin(alpha_v), 2*r_v, 0],
            [2*np.sin(alpha_h)*(1-d_h/f_h), 0, 2*((1-d_h/f_h)*r_lens_h + d_h)],
            [2*np.sin(alpha_v)*(1-d_v/f_v), 2*((1-d_v/f_v)*r_lens_v + d_v), 0]
        ])
    if (f_h == np.inf or f_v == np.inf
        or (d_h == 0 and d_v == 0) or (r_lens_h == 0 and r_lens_v == 0)):
        # If the length-sensing optical lever doesn't exist.
        c_align_inv[2:] = np.zeros_like(c_align_inv[2:])
        c_align_inv[:, 0] = np.zeros_like(c_align_inv[:, 0])
    elif d_h == 0 or r_lens_h == 0:
        # For vertical optical lever setup.
        c_align_inv[2, :] = np.zeros_like(c_align_inv[2, :])
    elif d_v == 0 or r_lens_v == 0:
        # For horizontal optical lever setup.
        c_align_inv[3, :] = np.zeros_like(c_align_inv[3, :])

    return np.linalg.pinv(c_align_inv).round(roundoff)


def c_rotation(phi_tilt, phi_len):
    r"""Returns the rotation transformation matrix for tilt and length QPDs.

    Parameters
    ----------
    phi_tilt: float
        Angle from the tilt-sensing QPD frame to the yaw-pitch frame.
    phi_len: float
        Angle from the length-sensing QPD frame to the yaw-pitch frame.

    Returns
    -------
    array
        Matrix that transform the QPD frames to the yaw-pitch frame.

    Notes
    -----
    See :math:`\mathbf{C}_\mathrm{rotation}` from [1]_

    References
    ----------
    .. [1]
        Tsang Terrence Tak Lun,
        Sensing Matrices for Optical Levers of the KAGRA Main Optics,
        https://github.com/terrencetec/kagra-optical-lever.
    """
    _c_rotation = np.array([
            [np.cos(phi_tilt), np.sin(phi_tilt), 0, 0],
            [-np.sin(phi_tilt), np.cos(phi_tilt), 0, 0],
            [0, 0, np.cos(phi_len), np.sin(phi_len)],
            [0, 0, -np.sin(phi_len), np.cos(phi_len)]
        ])
    return _c_rotation

def c_miscenter(delta_x, delta_y):
    """Returns the matrix for correcting miscentered optical lever beam.

    Parameters
    ----------
    delta_x: float
        Horizontal miscentering of the beam spot at the optics plane.
    delta_y: float
        Vertical miscentering of the beam spot at the optics plane.

    Returns
    -------
    array
        The miscentering correction matrix.
    """
    c_miscenter_inv = np.array([
            [1, delta_y, delta_x],
            [0, 1, 0],
            [0, 0, 1]
        ])
    return np.linalg.inv(c_miscenter_inv)



"""Utility functions for Dymanic Mode Decomposition."""
import numpy as np


def hankel(array, order):
    """Hankelize an array
    
    Parameters
    ----------
    array : array
        Array with rows as channels and columns as samples
    
    order : int
        The number of rows of the hankelized matrix.
    
    Returns
    -------
    hankel_array : array
        The hankelized array
    
    Examples
    --------
    .. code-block: python
        a = [1, 2, 3, 4, 5]
        hankel(a, 3)
        array([[1, 2, 3],
               [2, 3, 4],
               [3, 4, 5]])
    """
    hankel_array = np.zeros((order, len(array)-order+1))
    for i in range(order):
        hankel_array[i] = array[i:len(array)-order+1+i]
    return hankel_array


def auto_truncate(sigma, threshold=0.99):
    """Automatically get truncation value
    
    Parameters
    ----------
    sigma : array
        Singular values. Arranged from large to small values.
    threshold : float, optional
        Only include singular values so their sum is
        ``threshold`` of the original sum.
        Defaults 0.99.
        
    Returns
    -------
    truncation_value : int
        Amount of singular values that are required
        to reach the threshold sum.
    """
    sum_sigma = np.sum(sigma)
    for i in range(len(sigma)):
        truncation_value = i+1
        sum_truncated = np.sum(sigma[:truncation_value])
        if sum_truncated >= threshold*sum_sigma:
            break
    return truncation_value



"""Functions for designing post-regulator filters"""
import control
import numpy as np

import kontrol.logger
import kontrol.regulator.feedback
import kontrol.regulator.predefined


def post_low_pass(
        plant, regulator, post_filter=None,
        ignore_ugf_above=None, decades_after_ugf=1,
        phase_margin=45, f_start=None, f_step=1.1,
        low_pass=None, mtol=1e-6, small_number=1e-6,
        oscillatory=True, **kwargs):
    """Add low-pass filter after regulator.

    This function lowers/increase the the cutoff frequency
    of a low-pass filter until the phase margin at a
    dedicated ugf crosses the specified phase margin.
    Then, runs a bisection algorithm to poolish the cutoff
    frequency until the phase margin converges relative to
    the specified tolerance.

    Parameters
    ----------
    plant : TransferFunction
        The transfer function of the system that needs to be controlled.
    regulator : TransferFunction
        The regulator.
    post_filter : TransferFunction, optional
        Any post filters that will be applied on top of the regulator.
        Defaults None.
    ignore_ugf_above : float, optional
        Ignore unity gain frequencies higher than ``ignore_ugf_above`` (Hz).
        If not specified, defaults to 1 decade higher than the last UGF.
        This value can be overrided by the argument ``decades_after_ugf``
        Note that there's no guarantee that the UGF will be lower than this.
        The priority is to match the target phase margin.
        Defaults to None.
    decades_after_ugf : float, optional
        Set ignore_ugf_above some decades higher than the UGF of the OLTF
        ignore_ugf_above is None.
        Defaults to 1.
    phase_margin : float, optional,
        The target phase margin (Degrees).
        Defaults to 45.
    f_start : float, optional,
        The cutoff frequency to start iterating with.
        If not specified, defaults to some decades higher than the
        highest UGF. "Some decade" is set by ``decades_after_ugf``.
        Defaults None.
    f_step : float, optional,
        The gain that is used to multiply (or divide) the cutoff frequency
        during a coarse search.
        Defaults 1.1
    low_pass : func(cutoff, order) -> TransferFunction, optional
        The low-pass filter.
        If not specified, ``kontrol.regulator.predefined.lowpass()``
        with order 2 will be used.
        Defaults to None.
    mtol : float, optional
        Tolerance for convergence of phase margin.
        Defaults to 1e-6.
    small_number : float, optional
        A small number as a delta f to detect whether the gain is a rising or
        lowering edge at the unity gain frequency.
        Defaults to 1e-6.
    oscillatory : boolean, optional
        Use the first mode of the oscillatory system to evaluate the phase
        margins to avoid having UGFs at steep phase slopes.
        The benefit of using this is to have a conservative phase margin
        estimate. The phase response of the first mode
        is the lower bound of the phase response.
        If False, use the plant itself to calculate phase margins.
        If the plant does not contain any complex poles, this option will be
        overridden to False.
        Defaults True.
    **kwargs
        Keyword arguments passed to ``low_pass``.

    Returns
    -------
    TransferFunction
        The low-pass filter.
    """
    if oscillatory:
        wn, q, k = kontrol.regulator.feedback.mode_decomposition(plant)
        if len(wn) == 0:
            oscillatory = False
    if low_pass is None:
        low_pass = kontrol.regulator.predefined.low_pass
    if "order" not in kwargs.keys():
        kwargs["order"] = 2
    if post_filter is None:
        post_filter = control.tf([1], [1])

    regulator *= post_filter
    oltf = plant * regulator
    _, pms, _, _, ugfs, _ = control.stability_margins(oltf, returnall=True)
 
    if ignore_ugf_above is None:
        ignore_ugf_above = max(ugfs)/2/np.pi * 10**decades_after_ugf
    if f_start is None:
        f_start = max(ugfs)/2/np.pi * 10**decades_after_ugf

    ## Make initial guesses for the low-pass filter cutoff frequency.
    lower_edge_mask = (
        (abs(oltf(1j*(ugfs+ugfs*small_number)))
         < abs(oltf(1j*(ugfs-ugfs*small_number))))
    )  # Ignore UGFs with raising gain

    # Ignore UGFs higher than ignore_ugf_above.
    ignore_ugf_mask = ugfs/2/np.pi < ignore_ugf_above

    mask = lower_edge_mask * ignore_ugf_mask  #
    ugfs = ugfs[mask]
    if not oscillatory:
        pms = pms[mask]
    else:
        pms = []
        plant_one_mode = kontrol.regulator.feedback.mode_composition(
            wn[-1:], q[-1:], k[-1:])
        oltf_ = plant_one_mode * regulator
        # For each UGF, scale the OLTF one-mode plant such that the UGF matches
        # Then, find the phase margin using control.stability_margins.
        for ugf in ugfs:
            oltf_ *= 1/abs(oltf_(1j*ugf))  # Equalize the ugf
            _, pms_, _, _, ugfs_, _ =  control.stability_margins(
                oltf_, returnall=True)
            for i, ugf_ in enumerate(ugfs_):
                if i == 0:
                    ugf_phase_eval = ugf_
                    index_phase_eval = i
                else:
                    if abs(ugf_-ugf) < abs(ugf_phase_eval-ugf):
                        ugf_phase_eval = ugf_
                        index_phase_eval = i
            pms.append(pms_[index_phase_eval])          
        pms = np.array(pms)

    # Count phase margins that are already lower than
    # the specified phase margin and ignore them.
    n_pm_ignore = np.sum(pms < phase_margin)
    
    # Count number of phase margins
    n_pm = len(pms)

    if n_pm_ignore == n_pm:
        raise ValueError("All phase margins already below"
                         " specified phase margins."
                         " Can't implement low-pass filter.")
    if n_pm_ignore > 0:
        kontrol.logger.logger.warning(f"There are {n_pm_ignore} phase margins"
                                      " below the specified phase margin"
                                      " initially.")

    fc = f_start

    pm_was_lower = False
    pm_was_higher = False

    # Update 2022-01-17: Use the low-passed OLTF to evaluate UGF
    # but only use the first mode of the plant (if the plant is oscillatory)
    # to evaluate phase.
    
    # Start coarse searching
    while 1:
        oltf_lp = oltf * low_pass(fc, **kwargs)
        _, pms, _, _, ugfs, _ = control.stability_margins(
            oltf_lp, returnall=True)
        lower_edge_mask = ((abs(oltf_lp(1j*(ugfs+ugfs*small_number)))
                            < abs(oltf_lp(1j*(ugfs-ugfs*small_number)))))
        ignore_ugf_mask = ugfs/2/np.pi < ignore_ugf_above
        mask = lower_edge_mask * ignore_ugf_mask
        ugfs = ugfs[mask]
        if not oscillatory:
            pms = pms[mask]
        else:
            pms = []
            plant_one_mode = kontrol.regulator.feedback.mode_composition(
                wn[-1:], q[-1:], k[-1:])
            oltf_ = plant_one_mode * regulator * low_pass(fc, **kwargs)
            # For each UGF, scale the OLTF one-mode plant
            # such that the UGF matches.
            # Then, find the phase margin using control.stability_margins.
            for ugf in ugfs:
                oltf_ *= 1/abs(oltf_(1j*ugf))  # Equalize the ugf
                _, pms_, _, _, ugfs_, _ =  control.stability_margins(
                    oltf_, returnall=True)
                for i, ugf_ in enumerate(ugfs_):
                    if i == 0:
                        ugf_phase_eval = ugf_
                        index_phase_eval = i
                    else:
                        if abs(ugf_-ugf) < abs(ugf_phase_eval-ugf):
                            ugf_phase_eval = ugf_
                            index_phase_eval = i
                pms.append(pms_[index_phase_eval])          
            pms = np.array(pms)
        
        # Check number of phase margins and see if it changed. 
        # If it changed, 
        # recount the number of phase margins already lower than target.
        if len(pms) != n_pm:
            if n_pm_ignore > np.sum(pms < phase_margin):
                # In case the low-pass filter suppressed the problematic peak
                n_pm_ignore = np.sum(pms < phase_margin)
            n_pm = len(pms)
        
        # Find the minimum phase margin. (Don't count ignored ones.)
        pms_order = np.argsort(pms)
        min_pm_index = pms_order[n_pm_ignore]
        min_pm = pms[min_pm_index]
        if min_pm < phase_margin:
            f1 = fc
            pm_lower_bound = min_pm
            pm_was_lower = True
            fc *= f_step
        else:
            f2 = fc
            pm_upper_bound = min_pm
            pm_was_higher = True
            fc /= f_step
        
        # Phase margin crosses specification
        # memorize the two frequencies f1, f2 and break.
        if pm_was_lower and pm_was_higher:
            break

    # Runs bisection alogrithm using f1 and f2 as boundary.
    while 1:  # TODO Add maxiter
        fm = 10**((np.log10(f1) + np.log10(f2))/2)
        oltf_lp = oltf * low_pass(fm, **kwargs)
        _, pms, _, _, ugfs, _ = control.stability_margins(
            oltf_lp, returnall=True)
        lower_edge_mask = ((abs(oltf_lp(1j*(ugfs+ugfs*small_number)))
                           < abs(oltf_lp(1j*(ugfs-ugfs*small_number)))))
        ignore_ugf_mask = ugfs/2/np.pi < ignore_ugf_above
        mask = lower_edge_mask * ignore_ugf_mask
        ugfs = ugfs[mask]
        if not oscillatory:
            pms = pms[mask]
        else:
            pms = []
            plant_one_mode = kontrol.regulator.feedback.mode_composition(
                wn[-1:], q[-1:], k[-1:])
            oltf_ = plant_one_mode * regulator * low_pass(fm, **kwargs)

            # For each UGF, scale the OLTF one-mode plant
            # such that the UGF matches.
            # Then, find the phase margin using control.stability_margins.
            for ugf in ugfs:
                oltf_ *= 1/abs(oltf_(1j*ugf))  # Equalize the ugf
                _, pms_, _, _, ugfs_, _ =  control.stability_margins(
                    oltf_, returnall=True)
                for i, ugf_ in enumerate(ugfs_):
                    if i == 0:
                        ugf_phase_eval = ugf_
                        index_phase_eval = i
                    else:
                        if abs(ugf_-ugf) < abs(ugf_phase_eval-ugf):
                            ugf_phase_eval = ugf_
                            index_phase_eval = i
                pms.append(pms_[index_phase_eval])          
            #
            pms = np.array(pms)

        pms_order = np.argsort(pms)
        min_pm_index = pms_order[n_pm_ignore]
        min_pm = pms[min_pm_index]

        if (abs((min_pm - phase_margin)/phase_margin) <= mtol
                and min_pm > phase_margin):
            break
        if min_pm < phase_margin:
            f1 = fm
        else:
            f2 = fm
        if min_pm < pm_lower_bound or min_pm > pm_upper_bound:
            raise ValueError("Phase margin diverges during refinement.")
    
    return low_pass(fm, **kwargs)


def post_notch(
        plant, regulator=None, post_filter=None, target_gain=None,
        notch_peaks_above=None, phase_margin=45, notch=None, **kwargs):
    """Returns a list of notch filters that suppress resonance peaks.

    This functions finds the resonances peak of the plant/OLTF
    above certain frequencies and
    returns a list of notch filters that suppress
    these peaks to the target gains.

    Parameters
    ----------
    plant : TransferFunction
        The transfer function of the system that needs to be controlled.
    regulator : TransferFunction, optional
        The regulator.
        Defaults to None.
    post_filter : TransferFunction, optional
        Any post filters that will be applied on top of the regulator.
        Defaults None.
    target_gain : float, optional
        The target open-loop gain for the suppressed peak.
        To ensure a stable system, a value of less than 1 is recommended.
        If not specified, the notch will fully suppress the peak.
        Default None.
    phase_margin : float, optional
        The target phase margin.
        Defaults to 45.
    notch_peaks_above : float, optional
        Notch modes that has freqeuncies above ``notch_peaks_above``.
        If not specified, defaults to the highest unity gain frequency that
        is above ``phase_margin``
        Defaults to None. 
    notch : func(frequency, q, depth) -> TransferFunction, optional
        The notch filter.
        If not specified, ``kontrol.Notch()`` will be used.
        Defaults to None.
    **kwargs
        Keyword arguments passed to notch().

    Returns
    -------
    list of TransferFunction
        A list of notch filters.

    Notes
    -----
    This operation does not guarantee stability.
    It only finds resonances peaking out of the unity gain above some
    unity gain frequency and make notch filters to suppress them to
    a target gain level lower than the unity gain.
    The stability of the notched OLTF is not checked whatsoever.
    """
    if notch is None:
        notch = kontrol.Notch
    if regulator is None:
        regulator = control.tf([1], [1])
    if post_filter is None:
        post_filter = control.tf([1], [1])

    oltf = regulator * plant * post_filter
    oltf = oltf.minreal()
    _, pms, _, _, ugfs, _ = control.stability_margins(oltf, returnall=True)
    # mask out acceptable ugfs
    ufg_mask = pms <= phase_margin
    # Set the maximum one (masked) to be the target ugf.
    if notch_peaks_above is None:
        notch_peaks_above = max(ugfs[ufg_mask]/2/np.pi)

    wn, q, k = kontrol.regulator.feedback.mode_decomposition(oltf)
    # All poles in oltf above target frequency should be notched.
    fn = wn/2/np.pi
    notch_mask = fn > notch_peaks_above
    fn = fn[notch_mask]  # Resonance frequencies ordered from high to low.
    q = q[notch_mask]  # Quality factors
    k = k[notch_mask]  # DC gains
    gain_peak = q*k  # magnitude at the resonance
    
    # Choose modes that have peak higher than the DC gain.
    gain_mask = abs(gain_peak) > abs(k)
    fn = fn[gain_mask]
    q = q[gain_mask]
    k = k[gain_mask]
    gain_peak = gain_peak[gain_mask]

    # Notch peaks to the target gain or to their DC gain.
    notch_depth = np.zeros_like(fn)
    notch_q = np.zeros_like(fn)
    for i in range(len(notch_depth)):
        if target_gain is None:
            # Defaults to notch the peaks to the DC gain.
            notch_depth[i] = gain_peak[i]/k[i]
            notch_q[i] = 2*q[i]/notch_depth[i]
        elif target_gain < k[i]:
            # If the target gain is lower than the DC gain
            # Set target gain to DC gain instead.
            notch_depth[i] = gain_peak[i]/k[i]
            notch_q[i] = 2*q[i]/notch_depth[i]
        else:
            notch_depth[i] = gain_peak[i]/target_gain
            notch_q[i] = 2*target_gain/(k[i])

    notch_list = []
    for frequency, q, depth in zip(fn, notch_q, notch_depth):
        notch_list += [notch(frequency, q, depth)]

    return notch_list



"""Calibration library for calibrating sensors from sensors measurements"""
import numpy as np
import scipy.optimize

import kontrol.curvefit


def calibrate(xdata, ydata, method="linear", **kwargs):
    """Fit the measurement data and returns the slope of the fit.

    Parameters
    ----------
    xdata : array
        The independent variable, e.g. the displacement of the sensor.
    ydata : array
        The dependent variable, e.g. the sensor readout.
    method : str, optional
        The method of the fit.
        Choose from ["linear", "erf"].
       - "linear": use ``kontrol.sensact.calibration.calibrate_linear()``
       - "erf": use ``kontrol.sensact.calibration.calibrate_erf()``
    **kwargs :
        Keyword arguments passed to the calibration methods.

    Returns
    -------
    slope : float
        The slope of the fitted straight line.
    intercept : float
        The y-intercept of the fitted straight line.
    linear_range : float, optional
        The range of y where the sensor considered linear.
    """
    if method == "linear":
        return calibrate_linear(xdata, ydata, **kwargs)
    elif method == "erf":
        return calibrate_erf(xdata, ydata, **kwargs)
    else:
        raise ValueError("Invalid method. "
                         'Please choose from ["linear", "erf"].')


def calibrate_linear(
        xdata, ydata, nonlinearity=5, full_range=None, start_index=None,
        return_linear_range=False, return_model=False):
    """Fit a straight line to the data and returns the slope and intercept

    This functions recursively fits a straight line to chosen data points and
    terminates when no linear data point remains. It starts from 3 points
    closest to the middle of the full-range. After fitting a straight line
    to the 3 points, other data points that are within the linearity
    specification will be included to the dataset and this dataset will be
    use for the next fit. The process repeats until no data points can
    be added anymore.

    Parameters
    ----------
    xdata : array
        The independent variable, e.g. the displacement of the sensor.
    ydata : array
        The dependent variable, e.g. the sensor readout.
    nonlinearity : float, optional
        The specification of non-linearity (%).
        Defined by the maximum deviation of the full_range.
        Default 5.
    full_range : float, optional
        The full output range of the sensor.
        If not specified, it will be chosen to be
        ``max(ydata) - min(ydata)``
    start_index : int, optional
        The index of the data point to start with.
        If not specified, it will be chosen to be index of the ``ydata``
        closest to ``(max(ydata) + min(ydata))/2``.
    return_linear_range : boolean, optional
        If True, return the linear range of ydata.
    return_model : boolean, optional
        Return a kontrol.curvefit.model.Model object with the fitted
        parameters.
    
    Returns
    -------
    slope : float
        The slope of the fitted straight line.
    intercept : float
        The y-intercept of the fitted straight line.
    linear_range : float, optional
        The range of x where the sensor considered linear.
    model : kontrol.curvefit.model.Model
        The fitted model.
    """
    xdata = np.array(xdata)
    ydata = np.array(ydata)
    # Sort data
    sort_indexes = np.argsort(xdata)
    xdata = xdata[sort_indexes]
    ydata = ydata[sort_indexes]
    if full_range is None:
        full_range = max(ydata) - min(ydata)
    if start_index is None:
        mid_range = np.mean([max(ydata), min(ydata)])
        ## Find the element closest to the mid_range
        error = None
        for i in range(len(ydata)):
            if error is None:
                error = abs(ydata[i] - mid_range)
            if abs(ydata[i] - mid_range) < error:
                start_index = i
                error = abs(ydata[i] - mid_range)
    error_func = kontrol.curvefit.error_func.mse
    cost = kontrol.curvefit.Cost(error_func=error_func)
    model = kontrol.curvefit.model.StraightLine()
    optimizer = scipy.optimize.minimize
    curvefit = kontrol.curvefit.CurveFit()
    curvefit.model = model
    curvefit.cost = cost
    curvefit.optimizer = optimizer

    mask = np.zeros_like(xdata)==1
    for i in range(len(mask)):
        if (
            i == start_index or i-1 == start_index
            or i+1 == start_index):
            # Set the mask such the the middle points are true.
            mask[i] = True
    # Starting from 3 points in the middle, fit a straight line
    # and include those data points which fall within the linearity
    # specification, and repeats.
    while 1:
        curvefit.xdata = xdata[mask]
        curvefit.ydata = ydata[mask]
        x0_slope = ((curvefit.ydata[-1] - curvefit.ydata[0])
                    / (curvefit.xdata[-1] - curvefit.xdata[0]))
        x0_intersect = ydata[start_index] - x0_slope*xdata[start_index]
        x0 = [x0_slope, x0_intersect]
        optimizer_kwargs = {"x0": x0}
        curvefit.optimizer_kwargs = optimizer_kwargs
        curvefit.fit()
        line_fit = curvefit.model(xdata)
        nonlinearity_mask = abs(line_fit-ydata)/abs(full_range) * 100
        mask |= nonlinearity_mask < nonlinearity  # Include points
        # that are within the linearity specification.
        if len(xdata[mask]) == len(curvefit.xdata):
            # terminate when dataset didn't increase
            break
    
    slope = curvefit.model.slope
    intercept = curvefit.model.intercept
    
    returns = (slope, intercept)
    if return_linear_range:
        linear_range = max(curvefit.ydata) - min(curvefit.ydata)
        returns += (linear_range,)
    if return_model:
        returns += (curvefit.model,)

    return returns


def calibrate_erf(
        xdata, ydata, nonlinearity=5, return_linear_range=False,
        return_model=False):
    r"""Fit an error function and return the 1st-order slope and intercept.

    Parameters
    ----------
    xdata : array
        The independent variable, e.g. the displacement of the sensor.
    ydata : array
        The dependent variable, e.g. the sensor readout.
    nonlinearity : float, optional
        The specification of non-linearity (%).
        Defined by the maximum deviation of the full_range.
        Default 5.
    return_linear_range : boolean, optional
        If True, return the linear range of ydata.
    return_model : boolean, optional
        Return a kontrol.curvefit.model.Model object with the fitted
        parameters.
    
    Returns
    -------
    slope : float
        The slope of the fitted straight line.
    intercept : float
        The :math:`y`-intercept of the fitted straight line.
    linear_range : float, optional
        The range of x where the sensor considered linear.

    Notes
    -----
    Credits to Kouseki Miyo, the inventor of this method.

    We fit the following function
    
    .. math::
       f(x; a, m, x_0, y_0) = a\,\mathrm{erf}(m(x-x_0)) + y_0\,\,

    where :math:`a, m, x_0, y_0` are some parameters to be found.
    The :math:`\mathrm{erf}(x)` function is defined as

    .. math::
       \mathrm{erf}(x) = \frac{2}{\sqrt{\pi}}\int_0^x\,e^{-x^2}\,dx\,.

    If we taylor expand the exponential function and take the first-order
    approximation of the :math:`\mathrm{erf}(x)` function, we get

    .. math::
       \mathrm{erf}(x) \approx \frac{2am}{\sqrt{\pi}}(x-x_0) + y_0\,.

    So the (inverse) calibration factor is :math:`\frac{2am}{\sqrt{\pi}}`,
    and the :math:`y`-intercept is :math:`\frac{2am}{\sqrt{\pi}}x_0 + y_0`.
    """
    xdata = np.array(xdata)
    ydata = np.array(ydata)
    # Sort data
    sort_indexes = np.argsort(xdata)
    xdata = xdata[sort_indexes]
    ydata = ydata[sort_indexes]
    # Scale data for numerical stability
    xmean = np.mean(xdata)
    xdata -= xmean
    x_scale = max(abs(xdata))
    y_scale = max(abs(ydata))
    xdata /= x_scale
    ydata /= y_scale

    # Setup kontrol.curvefit
    error_func = kontrol.curvefit.error_func.mse
    cost = kontrol.curvefit.Cost(error_func=error_func)
    model = kontrol.curvefit.model.Erf()
    optimizer = scipy.optimize.minimize
    curvefit = kontrol.curvefit.CurveFit()
    curvefit.xdata = xdata
    curvefit.ydata = ydata
    curvefit.model = model
    curvefit.cost = cost
    curvefit.optimizer = optimizer

    # Guess initial parameters
    x0_amplitude = (max(ydata)-min(ydata)) / 2
    x0_slope = ((ydata[-1] - ydata[0]) / (xdata[-1] - xdata[0]))
    mid_range = np.mean([max(ydata), min(ydata)])
    ## Find the element closest to the mid_range
    error = None
    for i in range(len(ydata)):
        if error is None:
            error = abs(ydata[i] - mid_range)
        if abs(ydata[i] - mid_range) < error:
            start_index = i
            error = abs(ydata[i] - mid_range)
    x0_x_offset = xdata[start_index]
    x0_y_offset = ydata[start_index]
    x0 = [x0_amplitude, x0_slope, x0_x_offset, x0_y_offset]
    optimizer_kwargs = {"x0": x0}
    
    curvefit.optimizer_kwargs = optimizer_kwargs
    res = curvefit.fit()

    ## Unscaling
    curvefit.model.amplitude *= y_scale
    curvefit.model.slope /= x_scale
    curvefit.model.x_offset *= x_scale
    curvefit.model.x_offset += xmean
    curvefit.model.y_offset *= y_scale
    xdata *= x_scale
    xdata += xmean
    ydata *= y_scale

    a = curvefit.model.amplitude
    m = curvefit.model.slope
    x0 = curvefit.model.x_offset
    y0 = curvefit.model.y_offset
    slope = a*m*2/np.sqrt(np.pi)
    intercept = -slope*x0 + y0

    returns = (slope, intercept)
    if return_linear_range:
        y = kontrol.curvefit.model.StraightLine(args=[slope, intercept])
        full_range = max(curvefit.model(xdata)) - min(curvefit.model(xdata))
        nonlinearity_mask = (abs(y(xdata) - curvefit.model(xdata))
                             /full_range
                             * 100)
        mask = nonlinearity_mask < nonlinearity
        y_linear = ydata[mask]
        linear_range = max(y_linear) - min(y_linear)
        returns += (linear_range,)
    if return_model:
        returns += (curvefit.model,)

    return returns



"""Control regulators designs for oscillator-like systems"""
import control
import numpy as np

import kontrol.regulator.feedback
import kontrol.regulator.predefined


#FIXME find a better name
def pid(
    plant, regulator_type="PID", dcgain=None,
    integrator_ugf=None, integrator_time_constant=None,
    return_gain=False, **kwargs):
    """PID-like controller design for oscillator-like systems

    Parameters
    ----------
    plant : TransferFunction
        The transfer function of the system that needs to be controlled.
    regulator_type : str, optional
        The type of the contorl regulator.
        Choose from ``{"PID", "PD", "PI", "I", "D"}`` for
        proportional-integral-derivative, proportional-derivative,
        proportional-integral, or derivative (velocity) control respectively.
        Defaults to "PID".
    dcgain : float, optional
        The DC gain of the OLTF of the proportional control.
        If set to None, it will be set automatically
        depending on the type of the controller.
        If ``regulator_type=="PI"``, then dcgain will be set such that 
        the proportional control UGF matches that of the integral control.
        If ``regulator_type=="PD" or "PID"``, then the dcgain will be set
        such that it matches the UGF of the derivative control.
        Defaults to None.
    integrator_ugf : float, optional
        The unity gain frequency (Hz) of the integral control.
        This is the inverse of the integration time constant.
        If ``integrator_time_constant is not None``, then this value will be
        ignored.
        If set to None, it'll be set to match the UGF of the
        derivative control.
        For ``regulator_type=="I"``, this must be specified.
        Defaults to None.
    integrator_time_constant : float, optional,
        The integration time constant (s) for integral control.
        Setting this will override the ``integrator_ugf`` argument.
        Defaults to None.
    return_gain : boolean, optional
        Return the PID gain instead.
    
    Returns
    -------
    regulator : TransferFunction, optional
        The regulator.
        Return only if return_gain is ``False``.
    kp : float, optional
        Proportional gain.
        Return only if return_gain is ``True''.
    ki : float, optional
        Integral gain.
        Return only if return_gain is ``True``.
    kd : float, optional
        Derivative gain.
        Return only if return_gain is ``True``.
    """
    s = control.tf("s")
    kp = 0
    ki = 0
    kd = 0

    if regulator_type == "PID":
        kd = kontrol.regulator.feedback.critical_damping(plant, **kwargs)
        regulator = kontrol.regulator.predefined.pid(kd=kd)
        kp = kontrol.regulator.feedback.add_proportional_control(
            plant, regulator=regulator, dcgain=dcgain, **kwargs)
        # Commented line below so the open-loop gain don't go
        # below unity at low frequency.
        # regulator = kontrol.regulator.predefined.pid(kd=kd, kp=kp)
        ki = kontrol.regulator.feedback.add_integral_control(
            plant, regulator=regulator,
            integrator_ugf=integrator_ugf,
            integrator_time_constant=integrator_time_constant,
            **kwargs)
    elif regulator_type == "PD":
        kd = kontrol.regulator.feedback.critical_damping(plant, **kwargs)
        regulator = kontrol.regulator.predefined.pid(kd=kd)
        kp = kontrol.regulator.feedback.add_proportional_control(
            plant, regulator=regulator, dcgain=dcgain, **kwargs)
    elif regulator_type == "PI":
        kp = kontrol.regulator.feedback.add_proportional_control(
            plant, dcgain=dcgain, **kwargs)
        ki = kontrol.regulator.feedback.add_integral_control(
            plant, integrator_ugf=integrator_ugf,
            integrator_time_constant=integrator_time_constant)
    elif regulator_type == "I":
        ki = kontrol.regulator.feedback.add_integral_control(
            plant, integrator_ugf=integrator_ugf,
            integrator_time_constant=integrator_time_constant)
    elif regulator_type == "D":
        kd = kontrol.regulator.feedback.critical_damping(
            plant, **kwargs)
    else:
        raise ValueError('Invalid regulator_type. '
                         'Please select regulator_type from {"PID", "PD", '
                         '"PI", "D"}.')
    if return_gain:
        return kp, ki, kd
    else:
        regulator = kontrol.regulator.predefined.pid(kp=kp, ki=ki, kd=kd)
        return regulator



"""Noise models for empirical fitting.
"""
import numpy as np


def piecewise_noise(f, n0, exp=[0], fc=[0]):
    """Piecewise noise specified corner frequencies and exponents

    Parameters
    ----------
    f: list of int/float or numpy.ndarray
        The frequency axis of the noise.
    n0: int/float
        The noise level at 1 Hz with the first exponent.
    exp: list of int/float
        The list of exponents of each section of noise separated by the
        corner frequencies.
    fc: list of int/float
        The list of corner frequencies in increaing order. The length of
        fc must be 1 less then the length of exp

    Returns
    -------
    noise: numpy.ndarray
        The piecewise noise array.
    """
    if n0<0:
        n0 = -n0
    list(fc)
    if fc[-1] < np.inf:
        fc.append(np.inf)
    for i in range(len(fc)):
        if fc[i]<0:
            fc[i] = -fc[i]
    noise = np.zeros_like(f)
    fc_index = 0
    for i in range(len(f)):
        if f[i] >= fc[fc_index]:
            fc_index += 1
            n0 = n0 * fc[fc_index-1]**(exp[fc_index-1]-exp[fc_index])

        noise[i] = n0 * f[i]**exp[fc_index]
        # print(fc_index)
    return np.array(noise)


def lvdt_noise(f, n0=8e-3, fc=4.5, exp1=-0.5, exp2=0.):
    """LVDT noise model.

    Parameters
    ----------
    f: array
        Frequency axis.
    n0: float, optional.
        The noise level at 1 Hz with the first exponent. In um/rtHz.
        Defaults to the typical value 8e-3.
    fc: float, optional.
        Defaults to the typical value 4.5 Hz.
        The corner frequency at which the exponent changes.
    exp1: float, optional
        The exponent of the frequency dependency before the corner frequency.
        Defaults -0.5.
    exp2: float, optional
        The exponent of the frequency dependency after the corner frequency.
        Defaults 0.

    Returns
    -------
    noise: numpy.ndarray
        The LVDT noise frequency series.

    Notes
    -----
    The LVDT noise noise typically has a :math:`f^{-0.5}` dependency
    before the corner frequency and is flat after that.
    """
    return piecewise_noise(f, n0, exp=[exp1, exp2], fc=[fc])


def geophone_noise(f, n0=2e-6, fc=0.9, exp1=-3.5, exp2=-1.):
    """Geophone noise model.

    Parameters
    ----------
    f: array
        Frequency axis
    n0: float, optional
        The noise level at 1 Hz with the first exponent. In um/rtHz.
        Defaults to the typical value 2e-6.
    fc: float, optional
        The corner frequency at which the exponent changes.
        Defaults to the typical value 0.9.
    exp1: float, optional
        The exponent of the frequency dependency before the corner frequency.
        Defaults -3.5.
    exp2: float, optional
        The exponent of the frequency dependency after the corner frequency.
        Defaults -1.

    Returns
    -------
    noise: numpy.ndarray
        The geophone noise frequency series.

    Notes
    -----
    The geophone noise noise typically has a :math:`f^{-3.5}` dependency
    before the corner frequency and has a :math:`f^{-1}` dependency after that.
    """
    return piecewise_noise(f, n0, exp=[exp1, exp2], fc=[fc])



"""Some predefined complementary filters
"""
import control
import numpy as np
import scipy
import kontrol


def sekiguchi(coefs):
    r"""4th-order complementary filters specified by the blending frequencies.

    Parameters
    ----------
    coefs: float
        Blending frequency of the filter in [rad/s]

    Returns
    -------
    lpf: control.xferfcn.TransferFunction
        Complementary low pass-filter
    hpf: control.xferfcn.Transferfunction
        Complementary high pass-filter

    Notes
    -----
    4th-order complemetary filter used in Sekiguchi's
    thesis [1]_ whose high-pass is structured as:

    .. math::
        H = \frac{s^7 + 7w_bs^6 + 21w_b^2s^5 + 35w_b^3s^4}{(s+w_b)^7}

    where :math:`w_b` is the blending frequency in [rad/s].

    References
    ----------
    .. [1]
        T. Sekiguchi, A Study of Low Frequency Vibration Isolation System
        for Large Scale Gravitational Wave Detectors

    """

    blend_freq = coefs[0]
    _coefs = [blend_freq]*4
    # hpf_numerator = [
    #     1,
    #     7*blend_freq,
    #     21*blend_freq**2,
    #     35*blend_freq**3,
    #     0, 0, 0, 0,
    # ]
    # hpf = control.tf(hpf_numerator, [1]) * control.tf([1], [1, blend_freq])**7
    # lpf = 1 - hpf
    return modified_sekiguchi(_coefs)


def modified_sekiguchi(coefs):
    r"""Modified Sekiguchi Filter with guaranteed 4th-order high-pass.

    Parameters
    ----------
    coefs: list of (int or float) or numpy.ndarray
        4 coefficients defining the modified Sekiguchi filter.

    Returns
    -------
    lpf: control.xferfcn.TransferFunction
        Complementary low-pass filter
    hpf: control.xferfcn.Transferfunction
        Complementary high-pass filter

    Notes
    -----
        The modified Sekiguchi complemetary filter is structured as:

        .. math::
            H = \frac{s^7 + 7a_1s^6 + 21a_2^2s^5 + 35a_3^3s^4}{(s+a_4)^7}

        where :math:`a_1`, :math:`a_2`, :math:`a_3`, :math:`a_4` are some
        parameters that defines the filter.
    """

    a1, a2, a3, a4 = coefs
    hpf_numerator = [
        1,
        7 * a1,
        21 * a2**2,
        35 * a3**3,
        0, 0, 0, 0,
    ]
    hpf = control.tf(hpf_numerator, [1]) * control.tf([1], [1, a4])**7
    lpf = 1 - hpf
    return (lpf, hpf)


def generalized_sekiguchi(fb, order_low_pass, order_high_pass):
    """Generalized Sekiguchi Filter
    
    Parameters
    ----------
    fb : float
        Blending frequency in Hz.
    order_low_pass : int
        Order of roll-off of the low-pass filter
    order_high_pass : int
        Order of roll-off of the high-pass filter
    
    Returns
    -------
    lpf : kontrol.TransferFunction
        The low-pass filter
    hpf : kontrol.TransferFunction
        The high-pass filter
    """
    nl = order_low_pass
    nh = order_high_pass
    den = (scipy.special.comb(nl+nh-1, np.arange(0, nl+nh))
           *(2*np.pi*fb)**np.arange(0, nl+nh))
    lpf_num = den[nl:]
    lpf = kontrol.TransferFunction(lpf_num, den)
    return lpf, 1-lpf


def lucia(coefs):
    r"""Lucia Trozzo's complementary filter

    Parameters
    ----------
    coefs: list of float or numpy.ndarray of floats
        Takes 7 parameters, in specification order: :math:`p_1`, :math:`p_
        2`, :math:`z_1`, :math:`w_1`, :math:`q_1`, :math:`w_2`, :math:`q_2`
        . See notes for the meaning of the parameters.

    Returns
    -------
    lpf: control.xferfcn.TransferFunction
        Complementary low-pass filter
    hpf: control.xferfcn.Transferfunction
        Complementary high-pass filter
    Notes
    -----
    The modified Sekiguchi complemetary filter is structured as:

    .. math::
        H = K\frac{s^3(s+z_1)(s^2+w_1/q_1\,s+w_1^2)(s^2+w_2/q_2\,s+w_2^2)}
        {(s+p_1)^5(s+p_2)^3}

    where :math:`K` is a constant normalizing the filter at high frequency.
    """

    p1, p2, z1, w1, q1, w2, q2 = coefs
    hpf = control.tf([1], [1/p1, 1])**5 * control.tf([1], [1/p2, 1])**3
    hpf *= control.tf([1/z1, 1], [1])
    hpf *= control.tf([1, w1/q1, w1**2], [w1**2]) * control.tf([1, w2/q2, w2**2], [w2**2])
    hpf *= control.tf([1, 0],[1])**3
    hpf *= hpf.den[0][0][0]/hpf.num[0][0][0]
    lpf = 1 - hpf
    return (lpf, hpf)



"""Transfer function models for transfer function fitting."""
import control
import numpy as np

from .model import Model
import kontrol.core.math as math
import kontrol


class TransferFunctionModel(Model):
    r"""Transfer function model class defined by numerator and denominator

    Parameters
    ----------
    nzero : int
        Number of zeros.
    npole : int
        Number of poles.
    args : array or None, optional.
        The model parameters.
        Structured as follows:
        [b_n, b_n-1,..., b_1, b_0, a_m, a_m-1,..., a_1, a_0],
        where b and a are the coefficients of the numerator and denominator
        respectively, ordered from higher-order to lower-order.
        Defaults to None.

    Attributes
    ----------
    tf : kontrol.TransferFunction
        The last evaluted transfer function.

    Notes
    -----
    The transfer function model is defined as

    .. math::

       G(s, b_n, b_{n-1},..., b_1, b_0, a_m, a_{m-1},..., a_1, a_0)
       = \frac{\sum_{i=0}^{n} b_i s^i}{\sum_{j=0}^{m} a_j s^j}
    """
    def __init__(self, nzero, npole, args=None, log_args=False):
        """Constructor

        Parameters
        ----------
        nzero : int
            Number of zeros.
        npole : int
            Number of poles.
        args : array or None, optional.
            The model parameters.
            Structured as follows:
            [b_n, b_n-1,..., b_1, b_0, a_m, a_m-1,..., a_1, a_0],
            where b and a are the coefficients of the numerator and denominator
            respectively, ordered from higher-order to lower-order.
            Defaults to None.
        """
        super().__init__(args=args, log_args=log_args)
        self._nzero = None
        self._npole = None
        self.nzero = nzero
        self.npole = npole

    def _x2y(self, x, xunit="Hz"):
        """Transfer function frequency response.

        Parameters
        ----------
        x : array
            Independent variable in units specified by ``xunit``.
        xunit : str, optional.
            Unit of ``x``.
            Choose from ["Hz", "rad/s", "s"].
            Defaults to "Hz".

        Returns
        -------
        y : array
            Frequency response of the transfer function in complex values.
        """
        if len(self.num) + len(self.den) != self.nzero + self.npole + 2:
            raise ValueError("len(args) must be nzero+npole+2")
        num = self.num
        den = self.den
        s = _x2s(x, xunit)
        num_poly = math.polyval(num, s)
        den_poly = math.polyval(den, s)
        y = num_poly/den_poly
        return y

    @property
    def nzero(self):
        """Number of zeros"""
        return self._nzero

    @nzero.setter
    def nzero(self, _nzero):
        """nzero.setter"""
        self._nzero = _nzero

    @property
    def npole(self):
        """Number of zeros"""
        return self._npole

    @npole.setter
    def npole(self, _npole):
        """npole.setter"""
        self._npole = _npole

    @property
    def num(self):
        """Numerator array"""
        return self.args[:self.nzero+1]

    @property
    def den(self):
        """Denominator array"""
        return self.args[self.nzero+1:]

    @property
    def tf(self):
        """The Transfer Function object."""
        # if self.log_args:
        #     num = 10**self.num
        #     den = 10**self.den
        # else:
        num = self.num
        den = self.den
        return kontrol.TransferFunction(num, den)


class DampedOscillator(TransferFunctionModel):
    r"""Transfer function model for a damped oscillator.

    Parameters
    ----------
    args : array or None, optional.
        The model parameters with three numbers.
        Structured as follows:
        ``args = [k, fn, q]``,
        where ``k`` is the DC gain of the transfer function,
        ``fn`` is the resonance frequency in Hz, and ``q`` is the Q-factor.
        Defaults to None.

    Notes
    -----
    The model is definded as

    .. math::

       G(s; k, \omega_n, q) =
       k\frac{\omega_n^2}{s^2 + \frac{\omega_n}{q}s + \omega_n^2}

    where :math:`k` is the DC gain of the transfer function, :math:`\omega_n`
    is the resonance frequency of the oscillator, and :math:`q` is the
    Q-factor if the damped oscillator.
    """
    def __init__(self, args=None):
        """Constructor

        Parameters
        ----------
        args : array or None, optional.
            The model parameters with three numbers.
            Structured as follows:
            ``args = [k, fn, q]``,
            where ``k`` is the DC gain of the transfer function,
            ``fn`` is the resonance frequency in Hz, and ``q`` is the Q-factor.
            Defaults to None.
        """
        super().__init__(nzero=0, npole=2)
        self._args = None
        self.args = args

    @property
    def damped_oscillator_args(self):
        """The model parameters with three numbers [k, fn, q]"""
        return self._damped_oscillator_args

    @damped_oscillator_args.setter
    def damped_oscillator_args(self, _damped_oscillator_args):
        """damped_oscillator_args.setter"""
        self._damped_oscillator_args = _damped_oscillator_args

    # Overriding self.args in kontrol.curvefit.model.Model()
    @property
    def args(self):
        """Model parameters"""
        return self._args

    @args.setter
    def args(self, _args):
        """args.setter"""
        if _args is None:
            self._args = None
        else:
            if len(_args) != 3:
                raise ValueError("args must be in the format [k, fn, q].")
            self.damped_oscillator_args = _args
            k = self.gain
            fn = self.fn
            q = self.q
            wn = 2*np.pi*fn
            args = np.array([k*wn**2, 1, wn/q, wn**2])  # Convert to num, den.
            self._args = args

    @property
    def gain(self):
        """DC gain"""
        return self.damped_oscillator_args[0]

    @property
    def fn(self):
        """Resonance frequency"""
        return self.damped_oscillator_args[1]

    @property
    def q(self):
        """Q factor"""
        return self.damped_oscillator_args[2]


class CoupledOscillator(TransferFunctionModel):
    """"""
    def __init__(self):
        """"""
        pass
        # Taking a break.


class SimpleZPK(Model):
    r"""ZPK model with simple poles and zeros.

    Parameters
    ----------
    nzero : int
        Number of simple zeros.
    npole : int
        Number of simple poles.
    args : array, optional
         The model parameters defined as
         ``args = [z1, z2,..., p1, p2,..., k]`` 
         where ``z`` are the locations of the zeros in Hz,
         ``p`` are the locations of the pole in Hz, and
         ``k`` is the static gain,
    log_args : boolean, optional
        If true,
        model parameters passed to the model are assumed to be passed
        through a log10() function.
        So, when the real parameters will be assumed to be
        10**args instead.
        Defaults to False.
    
    Attributes
    ----------
    zero : array
        List of zeros.
    pole : array
        List of poles.
    gain : float
        Static gain.
    tf : kontrol.TranserFunction
        The transfer function representation of this ZPK model

    Notes
    -----
    The simple ZPK model is defined as

    .. math::

       G(s; z_1, z_2,..., p_1, p_2, ..., k)
       = k\frac{\prod_i(\frac{s}{2\pi z_i}+1)}{\prod_j(\frac{s}{2\pi p_j}+1)}
    """
    def __init__(self, nzero, npole, args=None, log_args=False):
        r"""Constructor.

        Parameters
        ----------
        nzero : int
            Number of simple zeros.
        npole : int
            Number of simple poles.
        args : array, optional
             The model parameters defined as
             ``args = [z1, z2,..., p1, p2,..., k]`` 
             where ``z`` are the locations of the zeros in Hz,
             ``p`` are the locations of the pole in Hz, and
             ``k`` is the static gain,
        log_args : boolean, optional
            If true,
            model parameters passed to the model are assumed to be passed
            through a log10() function.
            So, when the real parameters will be assumed to be
            10**args instead.
            Defaults to False.
        """
        self._nzero = None
        self._npole = None
        self.nzero = nzero
        self.npole = npole
        super().__init__(args=args, log_args=log_args)
    
    def _x2y(self, x, xunit="Hz"):
        """ZPK model frequency response."""
        s = _x2s(x, xunit)
        tf = np.ones_like(s) * self.gain 
        for i in range(len(self.zero)):
            tf *= s/(2*np.pi*self.zero[i]) + 1
        for i in range(len(self.pole)):
            tf /= s/(2*np.pi*self.pole[i]) + 1  
        return tf
                
    @property
    def nzero(self):
        """Number of simple zeros"""
        return self._nzero
    
    @nzero.setter
    def nzero(self, _nzero):
        """nzero.setter"""
        self._nzero = _nzero

    @property
    def npole(self):
        """Number of complex pole pairs"""
        return self._npole
    
    @npole.setter
    def npole(self, _npole):
        """npole.setter"""
        self._npole = _npole

    @property
    def args(self):
        """Model parameters in ZPK [z1, z2,..., p1, p2,..., k] format"""
        return self._args

    @args.setter
    def args(self, _args):
        """args.setter"""
        if _args is None:
            self._args = None
        elif len(_args) != self.nzero + self.npole + 1:
            raise ValueError("Length of argument must match nzero and npole.")
        else:
            if self.log_args:
                _args = 10**_args
            self._args = _args
    
    #TODO Maybe I should consider setter for zero, pole and gain.
    #These can be used to change the args as well but could be tedious.
    @property
    def zero(self):
        """List of zeros"""
        return self.args[:self.nzero]

    @property
    def pole(self):
        """List of poles"""
        return self.args[self.nzero:-1]

    @property
    def gain(self):
        """Static gain"""
        return self.args[-1]

    @property
    def tf(self):
        """Returns a TransferFunction object of this ZPK model"""
        s = control.tf("s")
        tf = control.tf([self.gain], [1])
        for i in range(len(self.zero)):
            tf *= s/(2*np.pi*self.zero[i]) + 1
        for i in range(len(self.pole)):
            tf /= s/(2*np.pi*self.pole[i]) + 1  
        return kontrol.TransferFunction(tf)


class ComplexZPK(Model):
    r"""ZPK model with complex poles and zeros.
    
    Parameters
    ----------
    nzero_pairs : int
        Number of complex zero pairs.
    npole_pairs : int
        Number of complex pole pairs.
    args : array, optional
         The model parameters defined as
         ``args = [f1, q1, f2, q2,..., fi, qi,..., fn, qn, k]``,
         where ``f`` are resonance frequencies of the complex
         zero/pole pairs, ``q`` are the quality factors, and
         ``k`` is the static gain, ``i`` is the number of complex zero
         pairs, and ``n-i`` is the number of of complex pole pairs.
    log_args : boolean, optional
        If true,
        model parameters passed to the model are assumed to be passed
        through a log10() function.
        So, when the real parameters will be assumed to be
        10**args instead.
        Defaults to False.

    Attributes
    ----------
    fn_zero : array
        List of resonance frequencies of the complex zeros.
    fn_pole : array
        List of resonance frequencies of the complex poles.
    q_zero : array
        List of Q-factors of the complex zeros.
    q_pole : array
        List of Q-factors of the complex poles.

    Notes
    -----
    The complex ZPK model is defined by:

    .. math::

       G(s; f_i, q_i,k)
       =k\frac{\prod_i(\frac{s^2}{(2\pi f_i)^2} + \frac{1}{2\pi f_i q_i}s + 1)}
       {\prod_j(\frac{s^2}{(2\pi f_j)^2} + \frac{1}{2\pi f_j q_j}s + 1)}

    """
    def __init__(
            self, nzero_pairs, npole_pairs, args=None, log_args=False):
        r"""Constructor.

        Parameters
        ----------
        nzero_pairs : int
            Number of complex zero pairs.
        npole_pairs : int
            Number of complex pole pairs.
        args : array, optional
             The model parameters defined as
             ``args = [f1, q1, f2, q2,..., fi, qi,..., fn, qn, k]``,
             where ``f`` are resonance frequencies of the complex
             zero/pole pairs, ``q`` are the quality factors, and
             ``k`` is the static gain, ``i`` is the number of complex zero
             pairs, and ``n-i`` is the number of of complex pole pairs.
        log_args : boolean, optional
            If true,
            model parameters passed to the model are assumed to be passed
            through a log10() function.
            So, when the real parameters will be assumed to be
            10**args instead.
            Defaults to False.

        Example
        -------
        with ``nzero_pairs = 1``, ``npole_pairs = 2``,
        args = [1, 2, 3, 4, 5, 6, 7] refers to a transfer function
        
        .. math::
           
           G(s) = 7\frac{\frac{1}{(1\times 2\pi)^2}s^2
           + \frac{1}{2\times 1 \ times 2\pi}s + 1}
           {\left(\frac{1}{(3\times 2\pi)^2}s^2
           + \frac{1}{4\times 3\times 2\pi}s + 1\right)
           \left(\frac{1}{(3\times 2\pi)^2}s^2
           + \frac{1}{4\times 3\times 2\pi}s + 1\right)}
        """
        self._nzero_pairs = None
        self._npole_pairs = None
        self._fn_zero = None
        self._fn_pole = None
        self._q_zero = None
        self._q_pole = None
        self._gain = None
        self.nzero_pairs = nzero_pairs
        self.npole_pairs = npole_pairs  # self.npole_pairs is not used.
        super().__init__(args=args, log_args=log_args)
    
    def _x2y(self, x, xunit="Hz"):
        """ZPK model (complex) frequency response."""
        s = _x2s(x, xunit)
        fn_zero = self.fn_zero
        q_zero = self.q_zero
        fn_pole = self.fn_pole
        q_pole = self.q_pole
        num = np.ones_like(s) * self.gain
        den = np.ones_like(s)
        for i in range(len(fn_zero)):
            num *= (1 / (2*np.pi*fn_zero[i])**2 * s**2
                    + 1 / (2*np.pi*fn_zero[i]*q_zero[i]) * s
                    + 1)
        for i in range(len(fn_pole)):
            den *= (1 / (2*np.pi*fn_pole[i])**2 * s**2
                    + 1 / (2*np.pi*fn_pole[i]*q_pole[i]) * s
                    + 1)
        return num/den
        
    @property
    def nzero_pairs(self):
        """Number of complex zero pairs"""
        return self._nzero_pairs
    
    @nzero_pairs.setter
    def nzero_pairs(self, _nzero_pairs):
        """nzero_pairs.setter"""
        self._nzero_pairs = _nzero_pairs

    @property
    def npole_pairs(self):
        """Number of complex pole pairs"""
        return self._npole_pairs
    
    @npole_pairs.setter
    def npole_pairs(self, _npole_pairs):
        """npole_pairs.setter"""
        self._npole_pairs = _npole_pairs
    
    @property
    def args(self):
        """Model parameters in ZPK [f1, q1, f2, q2,..., fn, qn, k] format"""
        return self._args

    @args.setter
    def args(self, _args):
        """args.setter"""
        if _args is None:
            self._args = None
        elif np.mod(len(_args), 2) != 1:
            raise ValueError("Length of argument must be odd number "
                             "and in the format "
                             "[f1, q1, f2, q2,..., fn, qn, k]")
        elif len(_args) != self.nzero_pairs*2 + self.npole_pairs*2 + 1:
            raise ValueError("Length of argument must match the specfied "
                             "nzero_pairs and npole_pairs.")
        else:
            if self.log_args:
                _args = 10**_args
            self._args = _args
    
    @property
    def fn_zero(self):
        """List of resonance frequencies of complex zeros."""
        return self.args[:int(self.nzero_pairs*2):2]

    @property
    def fn_pole(self):
        """List of resonance frequencies of complex poles."""
        return self.args[int(self.nzero_pairs*2):-1:2]

    @property
    def q_zero(self):
        """List of quality factors of the complex zeros"""
        return self.args[1:int(self.nzero_pairs*2):2]

    @property
    def q_pole(self):
        """List of quality factors of the complex poles"""
        return self.args[int(self.nzero_pairs*2)+1:-1:2]

    @property
    def gain(self):
        """Static gain."""
        return self.args[-1]
    
    @property
    def tf(self):
        """Returns a TransferFunction object of this ZPK model"""
        s = control.tf("s")
        fn_zero = self.fn_zero
        q_zero = self.q_zero
        fn_pole = self.fn_pole
        q_pole = self.q_pole
        k = self.gain
        num = control.tf([k], [1])
        den = control.tf([1], [1])
        for i in range(len(fn_zero)):
            num *= (1 / (2*np.pi*fn_zero[i])**2 * s**2
                    + 1 / (2*np.pi*fn_zero[i]*q_zero[i]) * s
                    + 1)
        for i in range(len(fn_pole)):
            den *= (1 / (2*np.pi*fn_pole[i])**2 * s**2
                    + 1 / (2*np.pi*fn_pole[i]*q_pole[i]) * s
                    + 1)
        return kontrol.TransferFunction(num/den)


#TODO add support for a generic ZPK model

        
def _x2s(x, xunit):
    """Converts the independent variable to the complex variable s.

    Parameters
    ----------
    x : array
        Independent variable in units specified by ``xunit``.
    xunit : str,
        Unit of ``x``.
        Choose from ["Hz", "rad/s", "s"].

    Returns
    -------
    s : array
        The complex variable.
    """
    if xunit == "Hz":
        s = 1j*2*np.pi*x
    elif xunit == "rad/s":
        s = 1j*x
    elif xunit == "s":
        s = x
    else:
        raise ValueError("Invalid specification for xunit."
                         "Please choose xunit from 'Hz', 'rad/s', or 's'.")
    return s





"""Test Kontrol's transfer function class. [Not really testing]
"""
import os

import control
import numpy as np

import kontrol
import kontrol.core.controlutils
import kontrol.regulator


def test_transfer_function_stablize():
    """Tests for kontrol.TransferFunction.stablize."""
    s = control.tf("s")
    tf_unstable = (s-2) / (s-1) / (s**2-s-1)
    kontrol_tf = kontrol.TransferFunction(tf_unstable)
    kontrol_tf.stabilize()
    positive_real_pole_exist = any(kontrol_tf.poles().real > 0)
    positive_real_zero_exist = any(kontrol_tf.zeros().real > 0)
    assert all([not positive_real_pole_exist, not positive_real_zero_exist]) 


def test_transfer_function_foton():
    tf = control.tf([1], [2, 3])
    kontrol_tf = kontrol.TransferFunction(tf)

    n_correct = "zpk([],[0.238732414637843],0.3333333333333333,\"n\")"
    kontrol_tf_foton_n = kontrol_tf.foton(root_location="n")
    f_correct = "zpk([],[-0.238732414637843],0.5,\"f\")"
    kontrol_tf_foton_f = kontrol_tf.foton(root_location="f")
    s_correct = "zpk([],[-1.5],0.5,\"s\")"
    kontrol_tf_foton_s = kontrol_tf.foton(root_location="s")
    rpoly_correct = "rpoly([1.0],[1.0;1.5],0.5)"
    kontrol_tf_foton_rpoly = kontrol_tf.foton(expression="rpoly")
    ## FIXME Make better comparison
    # assert all([
    #     kontrol_tf_foton_n==n_correct,
    #     kontrol_tf_foton_f==f_correct,
    #     kontrol_tf_foton_s==s_correct,
    #     kontrol_tf_foton_rpoly==rpoly_correct,])


def test_transfer_function_save_load():
    path = "tf_test.pkl"
    tf = control.ss2tf(control.rss(19))
    tf = kontrol.TransferFunction(tf)
    tf.save(path)
    # test exception
    try:
        tf.save(path, overwrite=False)
        raise
    except FileExistsError:
        pass

    tf2 = kontrol.load_transfer_function(path=path)
    os.remove(path)
    assert kontrol.core.controlutils.check_tf_equal(tf, tf2)

    # test for len(den) > len(num) case
    s = control.tf("s")
    tf3 = (s**2 + 2*s + 3) / (s + 3.14)
    tf3 = kontrol.TransferFunction(tf3)
    tf3.save(path)
    tf4 = kontrol.load_transfer_function(path)
    os.remove(path)
    assert kontrol.core.controlutils.check_tf_equal(tf3, tf4)

    # test file not found exception
    try:
        kontrol.load_transfer_function("File_that_probably_doesnt_exist.pkl")
        raise
    except FileNotFoundError:
        pass


def test_transfer_function_clean():
    """Tests for kontrol.TransferFunction.clean()
    """
    num = [1e-9, 1, 2, 3]
    den = [1e-9, 2, 3, 4]
    tf = control.tf(num, den)
    tf = kontrol.TransferFunction(tf)
    tf.clean()
    assert len(tf.num[0][0]) == 3
    assert len(tf.den[0][0]) == 3


def test_notch():
    """Tests for kontrol.transfer_function.notch.Notch class"""
    frequency = np.random.random()
    q = np.random.randint(1, 100)
    depth = np.random.randint(1, 100)
    depth_db = 20*np.log10(depth)

    # Test depth exception
    try:
        kontrol.Notch(frequency, q)
        raise
    except ValueError:
        pass

    notch = kontrol.Notch(frequency, q, depth)
    notch_db = kontrol.Notch(
        frequency, q, depth_db=depth_db)
    correct_notch = kontrol.regulator.predefined.notch(frequency, q, depth)
    assert kontrol.core.controlutils.check_tf_equal(correct_notch, notch)
    assert kontrol.core.controlutils.check_tf_equal(correct_notch, notch_db)

    # Test setters
    notch = kontrol.Notch(1, 2, 3)
    notch.frequency = frequency
    notch.q = q
    notch.depth = depth
    assert kontrol.core.controlutils.check_tf_equal(correct_notch, notch)

    # Test setter exceptions
    try:
        notch.frequency = -1
        raise
    except ValueError:
        pass
    try:
        notch.q = 0.1
        raise
    except ValueError:
        pass
    try:
        notch.depth = -1
        raise
    except ValueError:
        pass
    
    # Foton. Test for error only, won't check accuracy.
    notch.foton()



"""Tests for kontrol.frequency_series.frequency_series
"""
import numpy as np
import control

import kontrol
import kontrol.core.controlutils
import kontrol.frequency_series.noise_models


def test_frequency_series():
    np.random.seed(123)

    f = np.linspace(1e-1, 1e2, 100)
    s = 1j*2*np.pi*f

    ##  Lag filter at 1 & 10 Hz.
    def model(f, f1, f2):
        s = 1j*2*np.pi*f
        tf = (s/(2*np.pi*f2) + 1)/(s/(2*np.pi*f1) + 1)
        x = abs(tf)
        return x

    x = model(f, 1, 10)

    fs = kontrol.FrequencySeries(f=f, x=x)

    fs.fit_empirical(
        model=model, optimizer_kwargs={"options":{"maxfev":100000}})
    fs.fit_zpk(order=1, padding=True)
    fs.fit_tf()

    tf_correct = control.tf([1/(2*np.pi*10), 1], [1/(2*np.pi*1), 1])
    assert kontrol.core.controlutils.check_tf_equal(
        tf_correct, fs.tf, allclose_kwargs={"rtol":1e-2})


def test_noise_models():
    f = np.logspace(-2, 2, 1000)
    lvdt_noise = kontrol.frequency_series.noise_models.lvdt_noise(f=f)
    geophone_noise = kontrol.frequency_series.noise_models.geophone_noise(f=f)
    lvdt_noise_ = kontrol.frequency_series.noise_models.piecewise_noise(
        f=f, n0=8e-3, exp=[-0.5, 0], fc=[4.5])
    geophone_noise_ = kontrol.frequency_series.noise_models.piecewise_noise(
        f=f, n0=2e-6, exp=[-3.5, -1], fc=[0.9])
    assert all([
        np.allclose(lvdt_noise, lvdt_noise_),
        np.allclose(geophone_noise, geophone_noise_)])



"""Transfer function class.
Wrapper around control.TransferFunction
to provide custom functionality related to KAGRA.
"""
import os

import control
import numpy as np
import pickle

import kontrol.core.controlutils


class TransferFunction(control.TransferFunction):
    """Transfer function class

    Parameters
    ----------
    *args:
        Arguments passed to control.TransferFunction class.
    """
    def __init__(self, *args):
        """Initialize with a transfer function object

        Parameters
        ----------
        *args:
            Arguments passed to control.TransferFunction class.
        """
        super().__init__(*args)

    def lstrip(self, element, fc=None):
        """Remove zero or pole from the left.

        Parameters
        ----------
        element: str
            Element to be removed.
            Choose from ["any", "zero", "pole", "pair", "all"].
        fc: float or None, optional
            Cutoff frequency.
            If element is in ["any", "all"], remove any or all elements
            on the left of the cutoff.
            Defaults None.
        """
        # Work in progress
        pass

    def rstrip(self, element, fc=None):
        """Remove zero or pole from the right.

        Parameters
        ----------
        element: str
            Element to be removed.
            Choose from ["any", "zero", "pole", "pair", "all"].
        fc: float or None, optional
            Cutoff frequency.
            If element is in ["any", "all"], remove any or all elements
            on the left of the cutoff.
            Defaults None.
        """
        # Work in progress
        pass

    def clean(self, tol_order=5):
        """Remove numerator/denominator coefficients that are small outliers

        Parameters
        ----------
        tol_order : int, optional
            If the coefficient is ``tol_order`` order smaller than
            the rest of the coefficients, then this coefficient is an outlier.
            Defaults 5.
        """
        tf_cleaned = kontrol.core.controlutils.clean_tf3(
            self, tol_order=tol_order)
        super().__init__(tf_cleaned)


    def stabilize(self):
        """Convert unstable zeros and poles to stable ones.
        """
        stable_tf = kontrol.core.controlutils.convert_unstable_tf(self)
        super().__init__(stable_tf)

    def foton(
            self, expression="zpk", root_location="s", significant_figures=6,
            itol=1e-25, epsilon=1e-25):
        """Foton expression of this transfer function
        
        Calls kontrol.core.foton.tf2foton and returns a foton expression
        of this transfer function

        Parameters
        ----------
        expression : str, optional
            Format of the foton expression.
            Choose from ["zpk", "rpoly"].
            Defaults to "zpk".
        root_location : str, optional
            Root location of the zeros and poles for expression=="zpk".
            Choose from ["s", "f", "n"].
            "s": roots in s-plane, i.e. zpk([...], [...], ...,  "s").
            "f": roots in frequency plane, i.e. zpk([...], [,,,], ..., "f").
            "n": roots in frequency plane but negated and gains are normalized,
            i.e. real parts are positive zpk([...], [...], ..., "n").
            Defaults to "s".
        significant_figures : int, optional
            Number of significant figures to print out.
            Defaults to 6.
        itol : float, optional
            Treating complex roots as real roots if the ratio of
            the imaginary part and the real part is smaller than this tolerance
            Defaults to 1e-25.
        epsilon : float, optional
            Small number to add to denominator to prevent division error.
            Defaults to 1e-25.

        Returns
        -------
        foton_expression : str
            The foton expression in selected format.
        """
        ## TODO add MIMO support
        return kontrol.core.foton.tf2foton(
            tf=self, expression=expression, root_location=root_location,
            significant_figures=significant_figures,
            itol=itol, epsilon=epsilon)
    
    def save(self, path, overwrite=True):
        """Save the transfer function to a specified path.

        This functions extracts the numerator and denominator coefficients
        and puts it into a pandas DataFrame with keys {"num" and "den"}.
        Then it outputs to the specified path with a pickle format.

        Parameters
        ----------
        path : str
            The string of the path.
        overwrite : boolean, optional
            Overwrite if the file exists.
        """
        if not overwrite and os.path.exists(path):
            raise FileExistsError("File {} exists. "
                                  "Set overwrite option to True if you "
                                  "want to overwrite the file."
                                  "".format(path))
        with open(path, "wb") as f:
            pickle.dump(self, f)


def load_transfer_function(path):
    """Load a kontrol TransferFunction object from path.

    Parameters
    ----------
    path : str
        The path of the saved transfer function

    Returns
    -------
    TransferFunction
        The loaded TransferFunction object.
    """
    if not os.path.exists(path):
        raise FileNotFoundError("{} does not exist.".format(path))
    with open(path, "rb") as f:
        tf = pickle.load(f)
    return tf




import numpy as np

import kontrol.dmd.dmd
import kontrol.dmd.utils


def dmd_forecast(past, t_past, t_future,
                 order=None, truncation_value=None,
                 truncate_threshold=0.99, t_constant=None, i_constant=-1,
                 return_dmd=False):
    """Produce a short term forecast from past time series data using DMD.
    
    Parameters
    ----------
    past : array
        Past time series
    t_past : array
        Past time axis.
    t_future : array
        The time stamps of the future prediction.
        Preferably some time after t_past.
    order : int, optional
        The order of hankelization.
        Defaults to ``int(len(past)/2)``.
    truncation_value : int, optional
        The number of modes to be taken.
        If not provided, ``kontrol.dmd.auto_truncate()``
        will be used with the parameter ``truncate_threshold``
        to obtain the value.
    truncate_threshold : float, optional
        Truncate modes that has singular values less
        than this fraction of the total.
        This should within [0, 1], with 0 being all modes
        accepted and 1 being all modes truncated.
        Defaults to 0.99.
    t_constant : int, optional
        Time at which the constant vector is defined.
        Only make sense if it's a value in ``t_past``.
        Defaults t_past[-1].
    i_constant : int, optional
        The index of the column vector in the
        data matrix (snapshot) at ``t_constant``.
        Defaults -1.
    return_dmd : boolean, optional
        Return the DMD instance if True.
        Defaults to False.
    
    Returns
    -------
    forecast : array
        The forecast
    dmd : kontrol.dmd.DMD, optional
        The DMD instance used to produce the forecast.
        
    Examples
    --------
    
    .. code-block:: python
    
       t_past = np.linspace(0, 10, 1024) # Some time axis.
       t_future = np.linspace(10, 15, 512) # Forecast between t=[10, 15]. 
       time_series = ... # Some time series
       forecast = dmd_forecast(time_series, t_past, t_future)
    """
    if order is None:
        order = int(len(past)/2)

    # Hankelize the time series to create a data matrix.
    snapshot = kontrol.dmd.utils.hankel(array=past, order=order)
    
    
    # Create DMD instance
    dt = t_past[1]-t_past[0]
    dmd = kontrol.dmd.dmd.DMD(snapshot_1=snapshot, dt=dt)
    
    # Do an SVD of the snapshot and obtain a list of singular values.
    dmd.svd()
    
    # Obtain truncation value
    if truncation_value is None:
        truncation_value = kontrol.dmd.utils.auto_truncate(
            sigma=dmd.sigma, threshold=truncate_threshold)
    
    dmd.truncation_value = truncation_value
    
    # Run the DMD algorithm, get the DMD modes.
    dmd.run()
    
    # figure out how to handle t_predict and constants.
    if t_constant is None:
        t_constant = t_past[-1]
    prediction = dmd.predict(
        t=t_future, t_constant=t_constant, i_constant=i_constant)
    forecast = np.real(prediction[-1]) # Use last row for forecasting.

    if return_dmd:
        return forecast, dmd
    else:
        return forecast



"""KAGRA/LIGO Foton related utilities.
"""
import control
import numpy as np

import kontrol.core.controlutils
import kontrol.logger


def tf2foton(
        tf, expression="zpk", root_location="s", significant_figures=6,
        itol=1e-25, epsilon=1e-25):
    """Convert a single transfer function to foton expression.

    Parameters
    ----------
    tf : TransferFunction
        The transfer function object.
    expression: str, optional
        Format of the foton expression.
        Choose from ["zpk", "rpoly"].
        Defaults to "zpk".
    root_location : str, optional
        Root location of the zeros and poles for expression=="zpk".
        Choose from ["s", "f", "n"].
        "s": roots in s-plane, i.e. zpk([...], [...], ...,  "s").
        "f": roots in frequency plane, i.e. zpk([...], [,,,], ..., "f").
        "n": roots in frequency plane but negated and gains are normalized,
        i.e. real parts are positive zpk([...], [...], ..., "n").
        Defaults to "s".
    significant_figures : int, optional
        Number of significant figures to print out.
        Defaults to 6.
    itol : float, optional
        Treating complex roots as real roots if the ratio of
        the imaginary part and the real part is smaller than this tolerance
        Defaults to 1e-25.
    epsilon : float, optional
        Small number to add to denominator to prevent division error.
        Defaults to 1e-25.

    Returns
    -------
    foton_expression : str
        The foton expression in selected format.
    """
    if expression not in ["zpk", "rpoly"]:
        raise ValueError("expression {} not available."
                         "Please select expression from [\"zpk\", \"rpoly\"."
                         "".format(expression))
    ## Divide tf into tfs with less than 20 order.
    ## Do tf conversion.
    ## Stack the string
    foton_expression = ""
    tf_list = kontrol.core.controlutils.tf_order_split(tf, max_order=20)
    if len(tf_list) > 1:
        kontrol.logger.logger.warning("The transfer function has "
                                      "order higher than 20. This is not "
                                      "supported by KAGRA's Foton software. "
                                      "The Foton expression is splitted into "
                                      "multiple expressions with less order.")
    for tf_ in tf_list:
        if expression == "zpk":
            foton_expression += tf2zpk(
                tf_, root_location=root_location,
                significant_figures=significant_figures,
                itol=itol, epsilon=epsilon)
        elif expression == "rpoly":
            foton_expression += tf2rpoly(
                tf_, significant_figures=significant_figures)
        else:
            foton_expression += ""
            print("If you see this, contact maintainer.")
        foton_expression += "\n\n"
    foton_expression = foton_expression.rstrip("\n\n")

    return foton_expression


def tf2zpk(tf, root_location="s", significant_figures=6,
           itol=1e-25, epsilon=1e-25):
    """Convert a single transfer function to foton zpk expression.

    Parameters
    ----------
    tf : TransferFunction
        The transfer function object.
    root_location : str, optional
        Root location of the zeros and poles.
        Choose from ["s", "f", "n"].
        "s": roots in s-plane, i.e. zpk([...], [...], ...,  "s").
        "f": roots in frequency plane, i.e. zpk([...], [,,,], ..., "f").
        "n": roots in frequency plane but negated and gains are normalized,
        i.e. real parts are positive zpk([...], [...], ..., "n").
        Defaults to "s".
    significant_figures : int, optional
        Number of significant figures to print out.
        Defaults to 6.
    itol : float, optional
        Treating complex roots as real roots if the ratio of
        the imaginary part and the real part is smaller than this tolerance
        Defaults to 1e-25.
    epsilon : float, optional
        Small number to add to denominator to prevent division error.
        Defaults to 1e-25.

    Returns
    -------
    str
        The foton zpk expression in selected format.

    Notes
    -----
    Only works for transfer functions with less than 20 orders.
    """
    # if _order_gt(tf, 20):
    #     raise ValueError("Order of transfer function is not less than 20")
    if root_location not in ["s", "f", "n"]:
        raise ValueError("Select root_location from [\"s\", \"f\", \"n\"]")

    zeros = tf.zeros()
    poles = tf.poles()
    str_zeros = ""  # String of list of zeros (placeholder)
    str_poles = ""  # String of list of poles (placeholder)

    ## get zeros and poles.
    if root_location in ["f", "n"]:
        zeros /= 2*np.pi
        poles /= 2*np.pi
    if root_location == "n":
        zeros = -zeros.conjugate()
        poles = -poles.conjugate()

    ## get zeros and poles list, and sort.
    z_wn = np.sqrt(tf.zeros().real**2 + tf.zeros().imag**2)
    p_wn = np.sqrt(tf.poles().real**2 + tf.poles().imag**2)
    z_sort_arg = z_wn.argsort()
    p_sort_arg = p_wn.argsort()
    z_wn.sort()
    p_wn.sort()

    ## get gain
    gain = tf.minreal().num[0][0][0]
    if root_location in ["n"]:
        for wn in p_wn:
            if wn != 0:
                gain /= wn
            else:
                gain /= 2*np.pi
        for wn in z_wn:
            if wn != 0:
                gain *= wn
            else:
                gain *= 2*np.pi

    ## Convert to zpk expressing string
    for zero in zeros[z_sort_arg]:
        if abs(zero.imag)/abs(zero.real+epsilon) < itol:
            str_zeros += "{:.{}g}".format(zero.real, significant_figures)
        else:
            str_zeros += "{:.{sf}f}+i*{:.{sf}f}".format(
                zero.real, zero.imag, sf=significant_figures)
        str_zeros += ";"
    for pole in poles[p_sort_arg]:
        if abs(pole.imag)/abs(pole.real+epsilon) < itol:
            str_poles += "{:.{}g}".format(pole.real, significant_figures)
        else:
            str_poles += "{:.{sf}f}+i*{:.{sf}f}".format(
                pole.real, pole.imag, sf=significant_figures)
        str_poles += ";"
    str_zeros = str_zeros.rstrip(";")
    str_poles = str_poles.rstrip(";")
    zpk_expression = "zpk([{}],[{}],{:.{}g},\"{}\")".format(
        str_zeros, str_poles, gain, significant_figures, root_location)

    return zpk_expression


def tf2rpoly(tf, significant_figures=6):
    """Convert a transfer function to foton rpoly expression.

    Parameters
    ----------
    tf : TransferFunction
        The transfer function object
    significant_figures : int, optional
        Number of significant figures to print out.
        Defaults to 6.

    Returns
    -------
    str :
        Foton express in foton rpoly expression.

    Notes
    -----
    Only works for transfer functions with less than 20 orders.
    """
    if _order_gt(tf, 20):
        raise ValueError("Order of transfer function is not less than 20")

    num = tf.minreal().num[0][0]
    den = tf.minreal().den[0][0]
    str_num = ""  ## String of numerator coefficients
    str_den = ""  ## String of numerator coefficients
    gain = num[0]
    num /= num[0]

    for coef in num:
        str_num += "{:.{}g}".format(coef, significant_figures)
        str_num += ";"
    for coef in den:
        str_den += "{:.{}g}".format(coef, significant_figures)
        str_den += ";"
    str_num = str_num.rstrip(";")
    str_den = str_den.rstrip(";")
    rpoly_expression = "rpoly([{}],[{}],{})".format(str_num, str_den, gain)

    return rpoly_expression


def _order(tf):
    """Returns the number of coefficients in numerator and denominator

    Parameters
    ----------
    tf : TransferFunction
        The transfer function object

    Returns
    -------
    nnum : int
        Number of coefficients in numerator.
    nden : int
        Number of coefficients in denominator.
    """
    nnum = len(tf.minreal().num[0][0])
    nden = len(tf.minreal().den[0][0])
    return nnum, nden


def _order_gt(tf, order):
    """Returns true if transfer function order is greater than the specified.

    Parameters
    ----------
    tf : TransferFunction
        The transfer function object.
    order : int
        Order threshold.

    Returns
    -------
    boolean
        True if order(tf) > order, False otherwise.
    """
    nnum, nden = _order(tf)
    return max(nnum, nden) > order


def foton2tf(foton_string):
    """Convert a Foton string to TransferFunction
    
    Parameters
    ----------
    foton_string : str
        The Foton string, e.g. zpk([0], [1; 1], 1, "n").
    
    Returns
    -------
    tf : TransferFunction
        The transfer function.
    """
    # Categorize the string and call the relevant function.
    if "zpk" in foton_string:
        return zpk2tf(foton_string)
    elif "rpoly" in foton_string:
        return rpoly2tf(foton_string)
    else:
        raise ValueError("Foton string not recognized. "
                         "Only zpk and rpoly strings are supported.")


def zpk2tf(foton_string):
    """Convert a Foton ZPK string to TransferFunction
    
    Paramters
    ---------
    foton_string : str
        The Foton ZPK string, e.g. zpk([0], [1; 1], 1, "n").
    Returns
    -------
    tf : TransferFunction
        The transfer function.
    """
    foton_string = foton_string.replace(" ", "")
    foton_string = foton_string.lstrip("zpk(")
    foton_string = foton_string.rstrip(")")
    split = foton_string.split(",")
    if len(split) == 3:
        z_string, p_string, gain_string = split
        root_location = "s"
    else:
        z_string, p_string, gain_string, root_location = split

    def process_zp_string(zp_string):
        zp_string = zp_string.replace("[", "")
        zp_string = zp_string.replace("]", "")
        zp_string = zp_string.split(";")
        zp = []
        if zp_string == [""]:
            return np.array(zp)
        for string in zp_string:
            #split real and imaginary
            split = string.split("i*")
            if len(split) == 2:
                real_string, imag_string = split
                sign = real_string[-1]
            else:
                real_string, = split
                imag_string = "0"
                sign = "+"
            real_string = real_string.rstrip("+")
            real_string = real_string.rstrip("-")
            
            zp_complex = f"{real_string}{sign}{imag_string}j"
            zp.append(complex(zp_complex))
        zp = np.array(zp)
        return zp

    zeros = process_zp_string(z_string)
    poles = process_zp_string(p_string)
    gain = float(gain_string)

    root_location = root_location.replace("'", "")
    root_location = root_location.replace('"', "")
    # print(root_location)
    if root_location in "nf":
        # zeros poles are in Hz. Gain is DC.
        zeros *= 2*np.pi  # Convert to rad/s
        poles *= 2*np.pi
    #     print("nf")

    if root_location in "sf":
        # Negate for (s+z), (s+p) convention.
        zeros = -zeros
        poles = -poles

    tf = control.tf([1],[1])
    s = control.tf("s")

    for zero in zeros:
        if zero == 0:
            tf *= (s+zero)/(2*np.pi)
        elif zero.imag > 0 and zero.real != 0:
            wn = np.sqrt(zero.real**2 + zero.imag**2)
            q = wn / (2*zero.real)
            tf *= (s**2 + wn/q*s + wn**2) / wn**2
        elif zero.imag > 0 and zero.real == 0:
            wn = zero.imag
            tf *= (s**2 + wn**2)/wn**2
        elif zero.imag == 0 and zero.real != 0:
            tf *= (s+zero.real)/zero.real

    for pole in poles:
        if pole == 0:
            tf /= (s+pole)/(2*np.pi)
        elif pole.imag > 0 and pole.real != 0:
            wn = np.sqrt(pole.real**2 + pole.imag**2)
            q = wn / (2*pole.real)
            tf /= (s**2 + wn/q*s + wn**2) / wn**2
        elif pole.imag > 0 and pole.real == 0:
            wn = pole.imag
            tf /= (s**2 + wn**2)/wn**2
        elif pole.imag == 0 and pole.real != 0:
            tf /= (s+pole.real)/pole.real

    if root_location == "n":
        tf *= gain
    elif root_location in "sf":
        tf *= gain / (tf.num[0][0][0]/tf.den[0][0][0])
    
    return kontrol.TransferFunction(tf)


def rpoly2tf(foton_string):
    """Converts rpoly Foton strings to TransferFunction

    Parameters
    ----------
    foton_string : str
        The rpoly Foton string. E.g. rpoly([1; 2; 3], [2; 3; 4], 5)
    """
    foton_string = foton_string.lstrip("rpoly")
    foton_string = foton_string.replace(" ", "")
    foton_string = foton_string.replace("(", "")
    foton_string = foton_string.replace(")", "")
    num_string, den_string, gain_string = foton_string.split(",")
    def string2floatarray(string):
        """Convert a string of float array to array"""
        string = string.replace("[", "")
        string = string.replace("]", "")
        string = string.split(";")
        float_array = []
        if string == [""]:
            return np.array(float_array) 
        for value in string:
            float_array.append(float(value))
        return np.array(float_array)
    num = string2floatarray(num_string)
    den = string2floatarray(den_string)
    gain = float(gain_string)
    tf = control.tf(num, den)
    tf *= gain
    return kontrol.TransferFunction(tf)


def notch(frequency, q, depth, significant_figures=6):
    """Returns the foton expression of a notch filter.
    
    Parameters
    ----------
    frequency : float
        The notch frequency (Hz).
    q : float
        The quality factor.
    depth : float
        The depth of the notch filter (magnitude).
    significant_figures : int, optional
        Number of significant figures to print out.
        Defaults to 6.

    Returns
    -------
    str
        The foton representation of this notch filter.
    """
    depth_db = 20*np.log10(depth)
    expression = "notch({:.{sf}f},{:.{sf}f},{:.{sf}f})".format(
        frequency, q, depth_db, sf=significant_figures)
    return expression



"""Basic mathematical models"""
import numpy as np
import scipy.special

from .model import Model


class StraightLine(Model):
    r"""Simply a straight line.

    Parameters
    ----------
    args : array, optional
        The model parameters structued as:
        [slope, y-intercept].
        Defaults to None.

    Attributes
    ----------
    slope : float
        The slope of the line.
    intercept : float
        The y-intercept of the line.

    Notes
    -----
    The straight line is defined as

    .. math::
    
       y(x; m, c) = mx + c\,,

    where :math:`m` is the slope and :math:`c` is the y-intercept.
    """
    def __init__(self, args=None):
        """Constructor
        
        Parameters
        ----------
        args : array, optional
            The model parameters structued as:
            [slope, y-intercept].
            Defaults to None.
        """
        super().__init__(args=args)
        if self.args is None:
            self.args = np.zeros(2)

    @property
    def slope(self):
        """The slope of the line"""
        return self.args[0]
    
    @slope.setter
    def slope(self, _slope):
        """slope.setter"""
        self.args[0] = _slope
    
    @property
    def intercept(self):
        """The y-intercept of the line"""
        return self.args[1]


    @intercept.setter
    def intercept(self, _intercept):
        """intercept.setter"""
        self.args[1] = _intercept

    def _x2y(self, x, **kwargs):
        """y = mx + c"""
        y = self.slope*x + self.intercept
        return y


class Erf(Model):
    r"""The error function erf(x)

    Parameters
    ----------
    args : array, optional
        The model parameters structured as
        [amplitude, slope, x0, y0].
        Defaults to None.

    Notes
    -----
    The function is defined as
    
    .. math::

       f(x; a, m, x_0, y_0) = a\,\mathrm{erf}(m(x-x_0)) + y_0\,,
    
    where :math:`\mathrm{erf}(x)` is the error function [1]_,
    :math:`a` is the peak to peak amplitude, :math:`m` is the
    slope at the inflection point, :math:`x_0` is the :math:`x` offset
    , and :math:`y_0` is the :math:`y` offset.

    References
    ----------
    .. [1]
        https://en.wikipedia.org/wiki/Error_function
    """
    def __init__(self, args=None):
        """Constructor
        
        Parameters
        ----------
        args : array, optional
            The model parameters structured as
            [amplitude, slope, x0, y0].
            Defaults to None.
        """
        super().__init__(args=args)
        if self.args is None:
            self.args = np.zeros(4)

    def _x2y(self, x):
        """The error function"""
        a = self.amplitude
        m = self.slope
        x0 = self.x_offset
        y0 = self.y_offset
        return a*scipy.special.erf(m*(x-x0)) + y0

    @property
    def amplitude(self):
        """Peak to Peak amplitude of the error function."""
        return self.args[0]
    
    @amplitude.setter
    def amplitude(self, _amplitude):
        """amplitude.setter"""
        self.args[0] = _amplitude

    @property
    def slope(self):
        """Slope at the inflection point."""
        return self.args[1]
    
    @slope.setter
    def slope(self, _slope):
        """slope.setter"""
        self.args[1] = _slope

    @property
    def x_offset(self):
        """x offset"""
        return self.args[2]
    
    @x_offset.setter
    def x_offset(self, _x_offset):
        """x_offset.setter"""
        self.args[2] = _x_offset

    @property
    def y_offset(self):
        """y offset"""
        return self.args[3]
    
    @y_offset.setter
    def y_offset(self, _y_offset):
        """y_offset.setter"""
        self.args[3] = _y_offset







"""Base class for matrices, sensing and actuation matices.
"""
import numpy as np


class Matrix(np.ndarray):
    """Base class for matrices
    """
    def __new__(cls, matrix=None, *args, **kwargs):
        """Constructor

        Parameters
        ----------
        matrix: array
            The matrix/array.

        Returns
        -------
        obj: Matrix(numpy.ndarray)
            The matrix/array.
        """
        # print("Matrix __new__")
        self = np.asarray(matrix).view(cls)
        return self

    def __init__(self, *args, **kwargs):
        """Constructor
        """
        # print("Matrix __init__")
        pass


class SensingMatrix(Matrix):
    """Base class for sensing matrices
    """
    def __init__(self, matrix, coupling_matrix=None, *args, **kwargs):
        r"""Constructor

        Parameters
        ----------
        matrix: array
            The sensing matrix.
        coupling_matrix: array, optional
            The coupling matrix.
            Default None.

        Notes
        -----
        The coupling matrix has (i, j) elements as coupling ratios x_i/x_j.
        For example, consider the 2-sensor configuration:
        I have a coupled sensing readout :math:`x_{1,\mathrm{coupled}}`
        that reads :math:`x_{1,\mathrm{coupled}}=x_1 + 0.1x_2`.
        and, I have another coupled sensing readout
        :math:`x_{2,\mathrm{coupled}}` that reads
        :math:`x_{2,\mathrm{coupled}}=-0.2x_1 + x_2`.
        Then, the coupling matrix is

        .. math::
            \begin{bmatrix}
            1 & 0.1\\
            -0.2 & 1
            \end{bmatrix}.
        """
        # print("SensingMatrix __init__")
        # super().__init__(matrix)
        if coupling_matrix is not None:
            self.coupling_matrix = np.array(coupling_matrix)
        else:
            self.coupling_matrix = None

    def diagonalize(self, coupling_matrix=None):
        r"""Diagonalize the sensing matrix, given a coupling matrix.

        Parameters
        ----------
        coupling_matrix: array, optional.
            The coupling matrix.
            If None, self.coupling_matrix will be used.
            Default None.

        Returns
        -------
        array:
            The new sensing matrix.

        Notes
        -----
        The coupling matrix has (i, j) elements as coupling ratios x_i/x_j.
        For example, consider the 2-sensor configuration:
        I have a coupled sensing readout :math:`x_{1,\mathrm{coupled}}`
        that reads :math:`x_{1,\mathrm{coupled}}=x_1 + 0.1x_2`.
        and, I have another coupled sensing readout
        :math:`x_{2,\mathrm{coupled}}` that reads
        :math:`x_{2,\mathrm{coupled}}=-0.2x_1 + x_2`.
        Then, the coupling matrix is

        .. math::
            \begin{bmatrix}
            1 & 0.1\\
            -0.2 & 1
            \end{bmatrix}.
        """
        if coupling_matrix is not None:
            coupling_matrix = np.array(coupling_matrix)
            return np.linalg.inv(coupling_matrix) @ self
        elif self.coupling_matrix is not None:
            return np.linalg.inv(self.coupling_matrix) @ self
        else:
            raise ValueError("Coupling matrix is not specified")



"""Algorithmic designs for feedback control regulators.
"""
import control
import numpy as np

import kontrol


def critical_damping(plant, method="calculated", **kwargs):
    r"""Returns the critical damping derivative control gain.

    This functions calls
    ``kontrol.regulator.feedback.critical_damp_calculated()`` or
    ``kontrol.regulator.feedback.cricical_damp_optimized()`` and
    returns the derivative control gain.
    
    Parameters
    ----------
    plant : TransferFunction
        The transfer function representation of the system to be feedback
        controlled.
        The plant must contain at least one pair of complex poles.
    method : str, optional
        The method to be used for setting the gain.
        Choose from ["optimized", "calculated"].

        "optimized": the gain is optimized until the dominant complex
        pole pairs become two simple poles.

        "calculated": the gain is set to :math:`2/\omega_n/K_{DC}`,
        where :math:`\omega_n` is the resonance frequency in rad/s
        of the mode to be damped, and :math:`K_{DC}` is the sum of DC gains
        of the mode and that of the other high-frequency modes.
        
        Both method assumes that the plant has at least one pair of complex
        poles.

        Defaults to "calculated".
    **kwargs : dict
        Method specific keyword arguments.
        
        See:

        - "optimized": kontrol.regulator.feedback.critical_damp_optimized
        - "calculated": kontrol.regulator.feedback.critical_damp_calculated

    Returns
    -------
    kd : float
        The derivative gain for critically damping the
        dominant mode.
    """
    if _count_complex_poles(plant) == 0:
        raise ValueError("Plant must contain at least one pair of"
                         " complex poles")

    if method == "calculated":
        kd = critical_damp_calculated(plant, **kwargs)
    elif method == "optimized":
        kd = critical_damp_optimize(plant, **kwargs)
    else:
        raise ValueError('Invalid method. '
                         'Methods avavilable from ["optimized", "calculated"]')
    return kd 


def critical_damp_calculated(plant, nmode=1, **kwargs):
    r"""Find dominant mode and returns the approximate critical regulator.

    Parameters
    ----------
    plant : TransferFunction
        The transfer function representation of the system to be feedback
        controlled.
        The plant must contain at least one pair of complex poles.
    nmode : int, optional
        The ``nmode``th dominant mode to be damped.
        This number must be less than the number of modes in the plant.
        Defaults 1.

    Returns
    -------
    kd : float
        The derivative gain for critically damping the
        dominant mode.

    Notes
    -----
    The plant must contain at least one complex pole pair.
    The plant must have a non-zero DC gain.

    The critical regulator is approximated by

    .. math::
        K(s) \approx \frac{2}{\omega_n K_\mathrm{DC}}\,,

    where :math:`\omega_n` is the resonance frequency of the dominant mode
    and :math:`K_\mathrm{DC}` is the sum of DC gains of the mode
    and that of the other higher-frequency modes..
    """
    ## Decompose into frequencies, quality factors, and DC gains
    wn, q, k = kontrol.regulator.feedback.mode_decomposition(plant)
    peak_gain = q*k  # Gain at the resonances.

    # indexes of the modes from most dominant to least dominant.
    dominant_indexes = np.flip(np.argsort(peak_gain))
    # The index of the mode that we want to damp
    mode_index = dominant_indexes[nmode-1]

    # Sum all the DC gains of the modes that has higher 
    # or equal frequency than the mode (including the mode itself).
    dcgain = np.sum(k[:mode_index+1])
    dominant_wn = wn[mode_index]

    kd = 2/dominant_wn/dcgain  # Estimate critical damping gain.

    return kd 


def critical_damp_optimize(plant, gain_step=1.1, ktol=1e-6, **kwargs):
    r"""Optimize derivative damping gain and returns the critical regulator
    
    Parameters
    ---------- 
    plant : TransferFunction
        The transfer function representation of the system to be feedback
        controlled.
        The plant must contain at least one pair of complex poles.
    gain_step : float, optional
        The multiplicative factor of the gain for finding the gain
        upper bound.
        It must be greater than 1.
        Defaults to 1.1.
    ktol : float, optional
        The tolerance for the convergence condition.
        The convergence condition is (kd_max-kd_min)/kd_min > ktol.
        It must be greater than 0.
        Defaults to 1e-6.

    Returns
    -------
    kd : float
        The derivative gain for critically damping the
        dominant mode.
 
    Notes
    -----
    Update on 2021-12-04: Use carefully. It only critically damps
    the mode that has the highest peak when multiplied by an differentiator.
    Note for myself: only 2 complex poles can become simple poles for
    plants with second-order rolloff.

    Only works with plants that contain at least one complex pole pair.
    Works best with plants that only contain complex zeros/poles.
    If it returns unreasonably high gain, try lowering ``gain_step``.

    The algorithm goes as follows.

    1. Find the minimum damping gain ``k_min`` such that the open-loop
    transfer function ``k_min*s*plant`` has maxmimum gain
    at unity gain frequency.

    2. Iterate ``i``: ``k_i=k_min*i*gain_step`` for i=1,2,3...
    
    3. Terminate when ``1/(1+k_i*s*plant)`` has less complex pole pairs than
    the ``plant`` itself, i.e. one mode has been overdamped. Then, define
    ``k_max=k_i``.

    4. Iterate: Define ``k_mid`` as the logarithmic mean of
    ``k_max`` and ``k_min``. If ``1/(1+k_mid*s*plant)`` has less complex
    pole pairs than ``plant``, i.e. overdamps, then set ``k_max=k_mid``.
    Otherwise, set ``k_min=k_mid``.

    5. Terminate when ``(k_max - k_min)/k_min < ktol``, i.e. converges.
    """
    if gain_step <= 1:
        raise ValueError("gain_step must be greater than 1")
    if ktol <= 0:
        raise ValueError("ktol must be greater than 0")

    s = control.tf("s")
    f_pole = abs(plant.poles())/2/np.pi
    kd_min = 1/max(abs((s*plant)(1j*2*np.pi*f_pole)))

    k_min = kd_min * s
    oltf_min = k_min * plant

    n_complex_pole = _count_complex_poles(plant)

    # Find overdamp gain
    # Do so by adding gain until number of complex pole modes
    # reduced by more than the number of n_overdamp_modes
    kd = kd_min
    while 1:
        kd *= gain_step
        oltf = kd * s * plant
        sensitivity = plant / (1+oltf)
        damped_complex_poles = _count_complex_poles(sensitivity.minreal())
        if n_complex_pole  > damped_complex_poles:
            kd_max = kd
            break

    kd_mid = 10**((np.log10(kd_max)-np.log10(kd_min))/2 + np.log10(kd_min))

    # Keep tighting the bounds until convergence condition is met.
    while 1:
        kd_mid = 10**((np.log10(kd_max)-np.log10(kd_min))/2 + np.log10(kd_min))
        oltf_mid = kd_mid * s *plant
        sensitivity = plant / (1+oltf_mid)
        damped_complex_poles = _count_complex_poles(sensitivity.minreal())
        if n_complex_pole  > damped_complex_poles:
            # Mid gain still overdamps
            kd_max = kd_mid
        else:
            # Mid gain underdamps.
            kd_min = kd_mid
        
        if (kd_max-kd_min)/kd_min < ktol:
            # Convergence.
            kd_critical = kd_mid
            break

    kd = kd_mid

    return kd


def _count_complex_poles(plant):
    """Returns number of complex poles in a transfer function"""
    n_complex_pole = 0
    for p in plant.poles():
        if p.imag != 0:
            n_complex_pole += 1
    return n_complex_pole
    

def _find_dominant_mode(plant):
    """Returns the frequency (rad/s) of the dominant mode."""
    poles = plant.poles()
    dominant_wn = None
    for p in poles:
        if p.imag != 0:
            wn = abs(p)
            amp = abs(plant(1j*wn))
            if dominant_wn is None:
                dominant_wn = wn
            if dominant_wn is not None:
                if amp > abs(plant(1j*dominant_wn)):
                    dominant_wn = wn
    return dominant_wn


def _count_overdamp_modes(plant):
    """Count overdamped modes for critically damped dominant modes"""
    s = control.tf("s")
    n_overdamp_modes = 0
    dominant_wn = _find_dominant_mode(plant)
    spoles = (s*plant).pole()
    for p in spoles:
        if p.imag != 0:
            wn = abs(p)
            amp = abs((s*plant)(1j*wn))
            if amp > abs((s*plant)(1j*dominant_wn)):
                n_overdamp_modes += 1
    return n_overdamp_modes


# IDK where to put this function
# FIXME Put this function in a proper module. 
def mode_decomposition(plant):
    """Returns a list of single mode transfer functions

    Parameters
    ----------
    plant : TransferFunction
        The transfer function with at list one pair of complex poles.

    Returns
    -------
    wn : array
        Frequencies (rad/s).
    q : array
        Q factors.
    k : array
        Dcgains of the modes.
    """
    poles = plant.poles()
    complex_mask = poles.imag > 0  # Avoid duplication
    wn = abs(poles[complex_mask])  # Frequencies
    q = wn/(-2*poles[complex_mask].real)  # Q factors of the modes
    k = 1j * (plant(1j*wn)/q)  # DC gain of the modes
    # clean complex number
    sign = np.sign(k.real)
    k = sign * abs(k)
    return wn, q, k


def mode_composition(wn, q, k):
    """Create a plant composed of many modes.
    
    Parameters
    ----------
    wn : array
    Frequencies (rad/s).
    q : array
        Q factors.
    k : array
        Dcgains of the modes.
        
    Returns
    -------
    TransferFunction
        The composed plant.
    """
    if len(wn) != len(q) or len(wn) != len(k):
        raise ValueError("wn, q, k must have same length.")
    plant = kontrol.TransferFunction([0], [1])
    s = control.tf("s")
    for wn_, q_, k_ in zip(wn, q, k):
        plant = plant + k_ * wn_**2 / (s**2 + wn_/q_*s + wn_**2)
    return plant

    
def add_proportional_control(plant, regulator=None, dcgain=None, **kwargs):
    """Match and returns proportional gain.

    This function finds a proportional gain such that
    the proportional control UGF matches the first UGF
    of the specified regulator (typically a derivative control regulator).
    If ``dcgain`` is specified, the DC gain is matched instead.

    Parameters
    ----------
    plant : TransferFunction
        The transfer function representation of the system to be feedback
        controlled.
        The plant must contain at least one pair of complex poles.
    regulator : TransferFunction, optional
        The pre-regulator.
        If not specified, then ``dcgain`` must be specified.
        Defaults to None.
    dcgain : float, optional
        The desired DC gain of the open-loop transfer function.
        If not specified, the portional gain is tuned
        such that the portional control's first UGF matches that
        of the derivative control.
        Defaults to None.

    Returns
    -------
    kp : float
        The proportional control gain.
    """
    if dcgain is not None:
        kp = dcgain / plant.dcgain()
    elif regulator is not None:
        oltf = plant * regulator
        _, _, _, _, ugf, _ = control.stability_margins(
            oltf, returnall=True)
        kp = 1 / abs(plant(1j*min(ugf)))
    else:
        raise ValueError("At least one of regulator or dcgain must be "
                         "specified.")
    return kp


def add_integral_control(
        plant, regulator=None, integrator_ugf=None,
        integrator_time_constant=None, **kwargs):
    """Match and returns an integral gain.

    This function finds an integral gain such that
    the UGF of the integral control matches that of the specified
    regulator.
    If ``integrator_ugf`` or ``integrator_time_constant`` is specified
    instead, these will be matched instead.

    Parameters
    ----------
    plant : TransferFunction
        The transfer function representation of the system to be feedback
        controlled.
    regulator : TransferFunction, optional
        The pre-regulator
        Use ``kontrol.regulator.feedback.proportional_derivative()`` or
        ``kontrol.regulator.feedback.critical_damping()`` to make one
        for oscillator-like systems.
    integrator_ugf : float, optional
        The unity gain frequency (Hz) of the integral control.
        This is the inverse of the integration time constant.
        If ``integrator_time_constant is not None``, then this value will be
        ignored.
        If set to None, it'll be set to match the first UGF of the
        derivative control.
        Defaults to None.
    integrator_time_constant : float, optional,
        The integration time constant (s) for integral control.
        Setting this will override the ``integrator_ugf`` argument.
        Defaults to None.

    Returns
    -------
    ki : float
        The integral control gain.
    """
    s = control.tf("s")
    oltf_int = 1/s * plant.dcgain() 
    if integrator_time_constant is not None:
        integrator_ugf = 1/integrator_time_constant
        ki = 1 / abs(oltf_int(1j*2*np.pi*integrator_ugf))
    elif integrator_ugf is not None:
        ki = 1 / abs(oltf_int(1j*2*np.pi*integrator_ugf))
    elif regulator is not None:
        oltf = plant * regulator
        _, _, _, _, ugf, _ = control.stability_margins(
                    oltf, returnall=True)
        ki = 1 / abs(oltf_int(1j*min(ugf)))
    else:
        raise ValueError("At least one of regulator, integrator_ugf, or "
                         "integrator_time_constant must be specified.")
    return ki



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kontrol", # Replace with your own username
    version="1.1.0",
    author="TSANG Terrence Tak Lun",
    author_email="ttltsang@link.cuhk.edu.hk",
    description="KAGRA control python package",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/terrencetec/kontrol",
    packages=setuptools.find_packages(include=["kontrol", "kontrol.*"]),
    # packages=[
    #     "kontrol",
    # ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'control>=0.9'
    ],
    extra_requires={
        "ezca":["ezca"]
    },
)



"""Error functions for curve fitting
"""
import numpy as np


def mse(x1, x2, weight=None):
    """Mean square error

    Parameters
    ----------
    x1 : array
        Array
    x2 : array
        Array
    weight: array or None, optional
        weighting function.
        If None, defaults to np.ones_like(x1)

    Returns
    -------
    float
        Mean square error between arrays x1 and x2.
    """
    if weight is None:
        weight = np.ones_like(x1)
    return np.mean(np.square(np.subtract(x2, x1)*weight))


def log_mse(x1, x2, weight=None, small_multiplier=1e-6):
    """Logarithmic mean square error/Mean square log error

    Parameters
    ----------
    x1 : array
        Array
    x2 : array
        Array
    weight: array or None, optional
        weighting function.
        If None, defaults to `np.ones_like(x1)`
    small_multiplier: float, optional
        `x1` will be modified by `x1`+`small_multiplier`*`np.min(x2)` if
        0 exists in `x1`.
        Same goes to `x2`.
        If 0 both exists in `x1` and `x2`, raise.

    Returns
    -------
    float
        Logarithmic mean square error between arrays x1 and x2.
    """
    if 0 in x1 and 0 in x2:
        raise ValueError("zero value exists in both 'x1' and 'x2'"
                         "Can't take np.log10()")
    elif 0 in x1:
        x1 += small_multiplier * np.min(x2)
    elif 0 in x2:
        x2 += small_multiplier * np.min(x1)
    return mse(x1=np.log10(x1), x2=np.log10(x2), weight=weight)


def tf_error(tf1, tf2, weight=None, small_number=1e-6, norm=2):
    r"""Mean log error between two transfer functions frequency response

    Parameters
    ----------
    tf1 : array
        Frequency response of transfer function 1, in complex values.
    tf2 : array
        Frequency response of transfer function 2, in complex values.
    weight : array or None, optional
        Weighting function.
        Defaults None.
    small_number : float, optional
        A small number to be added in before evaluating np.log10 to
        prevent error.
        Defaults to 1e-6.
    norm : int, optional
        The type of norm used to estimate the error.
        Choose ``1`` for $\mathcal{L}_1$ norm and
        ``2`` for $\mathcal{L}_2$ norm.
        Other norms are are possible but not encouraged unless
        there's a specifiy reason.
        Defaults 2.

    Returns
    -------
    float
        Mean log error between two transfer function.

    Notes
    -----
    $\mathcal{L}_1$ norm is more robust while
    $\mathcal{L}_2$ norm is more susceptible to outliers, but fits
    data with large dynamic range well.
    """
    if weight is None:
        weight = np.ones_like(tf1, dtype=float)
    return np.mean(np.log10(abs(tf1-tf2)**norm + small_number) * weight)


def noise_error(noise1, noise2, weight=None, small_number=1e-6, norm=2):
    r"""Mean log error between two noise spectrums.

    Parameters
    ----------
    noise1 : array
        Noise spectrum 1.
    noise2 : array
        Noise spectrum 2.
    weight : array or None, optional
        Weighting function.
        Defaults None.
    small_number : float, optional
        A small number to be added in before evaluating np.log10 to
        prevent error.
        Defaults to 1e-6.
    norm : int, optional
        The type of norm used to estimate the error.
        Choose ``1`` for $\mathcal{L}_1$ norm and
        ``2`` for $\mathcal{L}_2$ norm.
        Other norms are are possible but not encouraged unless
        there's a specifiy reason.
        Defaults 2.

    Returns
    -------
    float
        The mean log error.

    Notes
    -----
    $\mathcal{L}_1$ norm is more robust while
    $\mathcal{L}_2$ norm is more susceptible to outliers, but fits
    data with large dynamic range well.
    """
    if weight is None:
        weight = np.ones_like(noise1, dtype=float)
    return np.mean((np.log10(abs(noise1))-np.log10(abs(noise2)))**norm*weight)


def spectrum_error(spectrum1, spectrum2,
                   weight=None, small_number=1e-6, norm=2):
    r"""Mean log error between two spectrums. Alias of noise_error().

    Parameters
    ----------
    spectrum1 : array
        Spectrum 1.
    spectrum2 : array
        Spectrum 2.
    weight : array or None, optional
        Weighting function.
        Defaults None.
    small_number : float, optional
        A small number to be added in before evaluating np.log10 to
        prevent error.
        Defaults to 1e-6.
    norm : int, optional
        The type of norm used to estimate the error.
        Choose ``1`` for $\mathcal{L}_1$ norm and
        ``2`` for $\mathcal{L}_2$ norm.
        Other norms are are possible but not encouraged unless
        there's a specifiy reason.
        Defaults 2.

    Returns
    -------
    float
        The mean log error.

    Notes
    -----
    $\mathcal{L}_1$ norm is more robust while
    $\mathcal{L}_2$ norm is more susceptible to outliers, but fits
    data with large dynamic range well.
    """
    return noise_error(spectrum1, spectrum2,
                       weight=weight, small_number=small_number, norm=norm)



"""Foton defined Notch filter.
"""
import kontrol.core.foton
import kontrol.regulator.predefined

from .transfer_function import TransferFunction


class Notch(TransferFunction):
    r"""Notch Filter Object
    
    Parameters
    ----------
    frequency : float
        The notch frequency (Hz).
    q : float
        The quality factor.
    depth : float, optional
        The depth of the notch filter (magnitude).
        If not specified, ``depth_db`` will be used.
        Defaults None.
    depth_db : float, optional
        The depth of the notch filter (decibel).
        If not specified, ``depth`` will be used instead.
        Defaults None.

    Attritubes
    ----------
    frequency : float
        The notch frequency (Hz).
    q : float
        The quality factor.
    depth : float
        The depth of the notch filter (magnitude).

    Notes
    -----
    The notch filter is defined by Foton, as
    
    .. math::
       
       N(s) = \frac{s^2 + (2\pi f_n)/(dQ/2)s + (2\pi f_n)^2}
       {s^2 + (2\pi f_n)/(Q/2)s + (2\pi f_n)^2}\,,

    where :math:`f_n` is the notch frequency, :math:`q` is the quality factor
    , and :math:`d` is the depth.    
    """
    def __init__(self, frequency, q, depth=None, depth_db=None):
        """Constructor

        Parameters
        ----------
        frequency : float
            The notch frequency (Hz).
        q : float
            The quality factor.
        depth : float, optional
            The depth of the notch filter (magnitude).
            If not specified, ``depth_db`` will be used.
            Defaults None.
        depth_db : float, optional
            The depth of the notch filter (decibel).
            If not specified, ``depth`` will be used instead.
            Defaults None.
        """
        if depth is None and depth_db is None:
            raise ValueError("Either depth or depth_db must be specified")
        if depth is None:
            depth = 10**(depth_db/20)
        _notch = kontrol.regulator.predefined.notch(
            frequency, q, depth, depth_db)
        super().__init__(_notch)
        self._frequency = None
        self._q = None
        self._depth = None
        self.frequency = frequency
        self.q = q
        self.depth = depth
        self.depth_db = depth_db

    @property
    def frequency(self):
        """The notch frequency (Hz)."""
        return self._frequency
    
    @frequency.setter
    def frequency(self, _frequency):
        """frequency.setter"""
        if _frequency < 0:
            raise ValueError("Frequency must be greater than 0 Hz")
        self._frequency = _frequency
        self._update_tf()

    @property
    def q(self):
        """The quality factor"""
        return self._q
    
    @q.setter
    def q(self, _q):
        """q.setter"""
        if _q < 0.5:
            raise ValueError("Q factor must be greater than 0.5")
        self._q = _q
        self._update_tf()

    @property
    def depth(self):
        """The depth of the notch filter (magnitude)."""
        return self._depth

    @depth.setter
    def depth(self, _depth):
        """depth.setter"""
        if _depth < 0:
            raise ValueError("Depth must be greater than zero")
        self._depth = _depth
        self._update_tf()

    def _update_tf(self):
        """Update the TransferFunction object"""
        if (self.frequency is not None 
                and self.q is not None 
                and self.depth is not None):
            frequency = self.frequency
            q = self.q
            depth = self.depth
            _notch = kontrol.regulator.predefined.notch(frequency, q, depth)
            super().__init__(_notch)

    def foton(self):
        """Foton expression of this notch filter."""
        return kontrol.core.foton.notch(self.frequency, self.q, self.depth)
